from .._utils import ContentCapture

async def role(role_name, hidden=False, _parser_context=None, **kwargs):
    ''' A chat role block.
    '''
    block_content = _parser_context['block_content']
    parser = _parser_context['parser']
    variable_stack = _parser_context['variable_stack']

    # capture the content of the block
    with ContentCapture(variable_stack, hidden) as new_content:
        
        # send the role-start special tokens
        new_content += parser.program.llm.role_start(role_name, **kwargs)

        # visit the block content
        new_content += await parser.visit(
            block_content,
            variable_stack,
            next_node=_parser_context["block_close_node"],
            prev_node=_parser_context["prev_node"],
            next_next_node=_parser_context["next_node"]
        )

        # send the role-end special tokens
        new_content += parser.program.llm.role_end(role_name)
role.is_block = True



example="""
{{#user~}}
Current summary:
The human asks what the AI thinks of artificial intelligence. The AI thinks artificial intelligence is a force for good.

New lines of conversation:
Human: Why do you think artificial intelligence is a force for good?
AI: Because artificial intelligence will help humans reach their full potential.

New summary:
{{~/user}}

{{#assistant~}}
The human asks what the AI thinks of artificial intelligence. The AI thinks artificial intelligence is a force for good because it will help humans reach their full potential.
{{~/assistant}}
"""

SUMMARIZER_TEMPLATE = """
{{#system~}}
You are a helpful and terse assistant.
{{~/system}}

{{#user~}}
Progressively summarize the lines of conversation provided, adding onto the previous summary returning a new summary.
{{~/user}}

{{#assistant~}}
Ok, I will do that. Let's do a practice round
{{~/assistant}}

"""+example+"""

{{#user~}}
That was great. Lets do another one.
Current summary:
{{summary}}

New lines of conversation:
{{ new_lines }}

New summary:
{{~/user}}

{{#assistant~}}
{{gen 'result' temperature=0}}
{{~/assistant}}
"""



async def if_(value, *, invert=False, _parser_context=None):
    ''' Standard if/else statement.

    Parameters
    ----------
    value : bool
        The value to check. If `True` then the first block will be executed, otherwise the second block
        (the one after the `{{else}}`) will be executed.
    invert : bool [DEPRECATED]
        If `True` then the value will be inverted before checking.
    '''
    block_content = _parser_context['block_content']
    variable_stack = _parser_context['variable_stack']
    parser = _parser_context['parser']
    
    assert len(block_content) % 2 == 1, "Unexpected number of blocks for `if` command: " + str(len(block_content))

    # parse the first block
    if invert:
        value = not value
    if value:
        return await parser.visit(block_content[0], variable_stack)

    # parse the rest of the blocks
    for i in range(1, len(block_content), 2):

        # elif block
        if block_content[i][0] == "elif":
            if parser.visit(block_content[i][1], variable_stack):
                return await parser.visit(block_content[i+1], variable_stack)

        # else block
        elif block_content[i][0] == "else":
            return await parser.visit(block_content[i+1], variable_stack)
        
        else:
            raise ValueError("Unexpected block content separator for `if` command: " + block_content[i].text)
    return ""
if_.is_block = True


import os
import pathlib
import openai
import yaml
import json
import asyncio
from tqdm import tqdm
from dotagent.fact_check.env_config import factool_env_config


# env
# openai.api_key = factool_env_config.openai_api_key


config = {
    'model_name': 'gpt-3.5-turbo',
    'max_tokens': 2000,
    'temperature': 0.0,
    'top_p': 1,
    'frequency_penalty': 0.0,
    'presence_penalty': 0.0,
    'n': 1
}


# Make api calls asynchronously
async def run_api(messages):
    async def single_run(message):
        output = openai.ChatCompletion.create(
            model=config['model_name'],
            messages=message,
            max_tokens=config['max_tokens'],
            temperature=config['temperature'],
            top_p=config['top_p'],
            frequency_penalty=config['frequency_penalty'],
            presence_penalty=config['presence_penalty'],
            n=config['n'],
        )
        return output.choices[0].message.content.strip()

    responses = [single_run(messages[index]) for index in range(len(messages))]
    return await asyncio.gather(*responses)



# Import data from scientific.json
scientific_list = []
with open("../datasets/scientific/scientific.json", "r") as f:
    data = json.load(f)
    for dict_data in data:
        cur_dict = {'dataset_name': 'scientific',
                    'question': dict_data["question"],
                    'factual_response': dict_data['factual_response']}
        scientific_list.append(cur_dict)

# Apply template prompt
with open("./prompts/claim_extraction.yaml") as f:
    data = yaml.load(f, Loader=yaml.FullLoader)
prompt = data['scientific']
messages_list = [
    [
        {"role": "system", "content": prompt['system']},
        {"role": "user", "content": prompt['user'].format(input=sample['factual_response'])},
    ]
    for sample in scientific_list
]

assert len(messages_list) == len(scientific_list), "The data length is different"

# Run the API to get the output
print("begin claims extraction...")
results = asyncio.run(run_api(messages_list))
for i in range(len(scientific_list)):
    scientific_list[i]["claims"] = results[i]

with open('../datasets/scientific/scientific_claims.json', 'w') as f:
    json.dump(scientific_list, f, indent=4)


"""
The scientific_claims.json file saved by the above code may have format problems, here are some adjustments
"""
with open("../datasets/scientific/scientific_claims.json", "r") as f:
    data = json.load(f)
    for data_i in tqdm(data, total=len(data)):
        try:
            data_i["claims"] = json.loads(data_i["claims"].strip())
        except:
            print(data_i["claims"])
            continue
with open("../datasets/scientific/scientific_claims.json", "w") as f:
    json.dump(data, f, indent=4)



from abc import ABC, abstractmethod


class BaseImageLlm(ABC):
    @abstractmethod
    def get_image_model(self):
        pass

    @abstractmethod
    def generate_image(self, prompt: str, size: int = 512, num: int = 2):
        pass


#New code implementation for the following api different from langchain approach

"""Util that calls SceneXplain.

In order to set this up, you need API key for the SceneXplain API.
You can obtain a key by following the steps below.
- Sign up for a free account at https://scenex.jina.ai/.
- Navigate to the API Access page (https://scenex.jina.ai/api) and create a new API key.
"""
from typing import Dict

import requests
from pydantic import BaseModel, BaseSettings, Field, root_validator

from dotagent.utils import get_from_dict_or_env
import base64
import http
import json


def _image_to_data_uri(file_path):
        with open(file_path, "rb") as image_file:
            encoded_image = base64.b64encode(image_file.read()).decode("utf-8")
            return f"data:image/jpeg;base64,{encoded_image}"

class SceneXplainAPIWrapper(BaseSettings, BaseModel):
    """Wrapper for SceneXplain API.

    In order to set this up, you need API key for the SceneXplain API.
    You can obtain a key by following the steps below.
    - Sign up for a free account at https://scenex.jina.ai/.
    - Navigate to the API Access page (https://scenex.jina.ai/api)
      and create a new API key.
    """

    scenex_api_key: str = Field(..., env="SCENEX_API_KEY")
    scenex_api_url: str = (
        "us-central1-causal-diffusion.cloudfunctions.net"
    )

    def _describe_image(self, image: str) -> str:
        local_image_path = image
        data = {
            "data": [
                {"image": _image_to_data_uri(local_image_path), "features": []},
            ]
        }

        headers = {
            "x-api-key": f"token {self.scenex_api_key}",
            "content-type": "application/json",
        }

        connection = http.client.HTTPSConnection("us-central1-causal-diffusion.cloudfunctions.net")
        connection.request("POST", "/describe", json.dumps(data), headers)
        response = connection.getresponse()
        response_data = response.read().decode("utf-8")
        response_data = json.loads(response_data)
        output = response_data["result"][0]["text"]
        connection.close()
        return output

    @root_validator(pre=True)
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that api key exists in environment."""
        scenex_api_key = get_from_dict_or_env(
            values, "scenex_api_key", "SCENEX_API_KEY"
        )
        values["scenex_api_key"] = scenex_api_key

        return values

    def run(self, image: str) -> str:
        """Run SceneXplain image explainer."""
        description = self._describe_image(image)
        if not description:
            return "No description found."

        return description
    



from typing import Any, Dict
import asyncio
import re
import json
from dotagent import compiler
from dotagent.caches import DiskCache

class LLMMeta(type):
    def __init__(cls, *args, **kwargs):
        cls._cache = None
    @property
    def cache(cls):
        if cls._cache is None:
            cls._cache = DiskCache(cls.llm_name)
        return cls._cache
    @cache.setter
    def cache(cls, value):
        cls._cache = value

class LLM(metaclass=LLMMeta):
    cache_version = 1
    default_system_prompt = "You are a helpful assistant."
    llm_name: str = "unknown"

    def __init__(self):
        self.chat_mode = False  # by default models are not in role-based chat mode
        self.model_name = "unknown"

        # these should all start with the @ symbol and are variables programs can use when running with this LLM
        self.tool_def = compiler("""
# Tools

{{#if len(functions) > 0~}}
## functions

namespace functions {

{{#each functions item_name="function"~}}
// {{function.description}}
type {{function.name}} = (_: {
{{~#each function.parameters.properties}}
{{#if contains(this, "description")}}// {{this.description}}
{{/if~}}
{{@key}}{{#unless contains(function.parameters.required, @key)}}?{{/unless}}: {{#if contains(this, "enum")}}{{#each this.enum}}"{{this}}"{{#unless @last}} | {{/unless}}{{/each}}{{else}}{{this.type}}{{/if}}{{#unless @last}},{{/unless}}
{{~/each}}
}) => any;

{{/each~}}
} // namespace functions
{{~/if~}}""", functions=[])
        self.function_call_stop_regex = r"\n?\n?```typescript\nfunctions.[^\(]+\(.*?\)```"

    def extract_function_call(self, text):
        m = re.match(r"\n?\n?```typescript\nfunctions.([^\(]+)\((.*?)\)```", text, re.DOTALL)

        if m:
            return CallableAnswer(m.group(1), m.group(2))

    def __call__(self, *args, asynchronous=False, **kwargs):
        """Creates a session and calls the LLM with the given arguments.

        Note that this is a convenience wrapper so you don't have to call session(),
        for higher performance across multiple calls, use a session directly.
        """
        with self.session(asynchronous=asynchronous) as s:
            out = s(*args, **kwargs)
        return out

    def __getitem__(self, key):
        """Gets an attribute from the LLM."""
        return getattr(self, key)

    def session(self, asynchronous=False):
        """Creates a session for the LLM.

        This implementation is meant to be overridden by subclasses.
        """

        if asynchronous:
            return LLMSession(self)
        else:
            return SyncSession(LLMSession(self))

    def encode(self, string, **kwargs):
        return self._tokenizer.encode(string, **kwargs)

    def decode(self, tokens, **kwargs):
        return self._tokenizer.decode(tokens, **kwargs)
    
    def id_to_token(self, id):
        return self.decode([id])

    def token_to_id(self, token):
        return self.encode(token)[0]
    
    # allow for caches to be get and set on the object as well as the class
    @property
    def cache(self):
        if self._cache is not None:
            return self._cache
        else:
            return self.__class__.cache
    @cache.setter
    def cache(self, value):
        self._cache = value


class LLMSession:
    def __init__(self, llm):
        self.llm = llm
        self._call_counts = {} # tracks the number of repeated identical calls to the LLM with non-zero temperature

    def __enter__(self):
        return self

    async def __call__(self, *args, **kwargs):
        return self.llm(*args, **kwargs)

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def _gen_key(self, args_dict):
        del args_dict["self"]  # skip the "self" arg
        return "_---_".join([str(v) for v in ([args_dict[k] for k in args_dict] + [self.llm.model_name, self.llm.__class__.__name__, self.llm.cache_version])])

    def _cache_params(self, args_dict) -> Dict[str, Any]:
        """get the parameters for generating the cache key"""
        key = self._gen_key(args_dict)
        # if we have non-zero temperature we include the call count in the cache key
        if args_dict.get("temperature", 0) > 0:
            args_dict["call_count"] = self._call_counts.get(key, 0)

            # increment the call count
            self._call_counts[key] = args_dict["call_count"] + 1
        args_dict["model_name"] = self.llm.model_name
        args_dict["cache_version"] = self.llm.cache_version
        args_dict["class_name"] = self.llm.__class__.__name__

        return args_dict


class SyncSession:
    def __init__(self, session):
        self._session = session

    def __enter__(self):
        self._session.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        return self._session.__exit__(exc_type, exc_value, traceback)

    def __call__(self, *args, **kwargs):
        return asyncio.get_event_loop().run_until_complete(
            self._session.__call__(*args, **kwargs)
        )

class CallableAnswer:
    def __init__(self, name, args_string, function=None):
        self.__name__ = name
        self.args_string = args_string

    def __call__(self, *args, **kwargs):
        if self._function is None:
            raise NotImplementedError(f"Answer {self.__name__} has no function defined")
        return self._function(*args, **self.__kwdefaults__, **kwargs)
    
    @property
    def __kwdefaults__(self):
        """We build this lazily in case the user wants to handle validation errors themselves."""
        return json.loads(self.args_string)

    def __repr__(self):
        return f"CallableAnswer(__name__={self.__name__}, __kwdefaults__={self.__kwdefaults__})"



import json
import math
import os
from typing import List, Dict
import yaml
import pdb

from dotagent.fact_check.math.tool import python_executor
from dotagent.fact_check.utils.base.pipeline import pipeline

class math_pipeline(pipeline):
    def __init__(self, foundation_model):
        super().__init__('math', foundation_model)

        self.tool = python_executor()

        with open(os.path.join(self.prompts_path, "claim_extraction.yaml"), 'r') as file:
            data = yaml.load(file, Loader=yaml.FullLoader)
        self.claim_prompt = data['math']

        with open(os.path.join(self.prompts_path, 'query_generation.yaml'), 'r') as file:
            data = yaml.load(file, Loader=yaml.FullLoader)
        self.query_prompt = data['math']

    def _verification(self, exec_results):
        classification_results = [True for _ in range(len(exec_results))]
        for i in range(len(exec_results)):
            if exec_results[i] is not None and 'False' in exec_results[i]:
                classification_results[i] = False
        
        return classification_results

    async def _claim_extraction(self, samples):
        messages_list = [
            [
                {"role": "system", "content": self.claim_prompt['system']},
                {"role": "user", "content": self.claim_prompt['user'].format(input_question=sample['prompt'], input_solution=sample['response'])},
            ]
            for sample in samples
        ]
        return await self.chat.async_run(messages_list, List)
    
    async def _query_generation(self, claims):
        messages_list = [
            [
                {"role": "system", "content": self.query_prompt['system']},
                {"role": "user", "content": self.query_prompt['user'].format(math_calculation=claim['math_calculation'], calculated_answer=claim['calculated_answer'])},
            ]
            for claim in claims
        ]
        return await self.chat.async_run(messages_list, Dict)
    
    async def run_with_tool_live(self, samples):
        claims_in_responses = await self._claim_extraction(samples)
        queries_in_responses = []
        exec_results_in_responses = []
        verifications_in_responses = []
        for claims_in_response in claims_in_responses:
            queries = await self._query_generation(claims_in_response)
            queries_in_responses.append(queries)
            exec_results = []
            for query in queries:
                try:
                    exec_results.append(self.tool.run(query['python_snippet']))
                except:
                    exec_results.append('None')
            exec_results_in_responses.append(exec_results)
            verifications = self._verification(exec_results)
            verifications_in_responses.append(verifications)

        return claims_in_responses, queries_in_responses, exec_results_in_responses, verifications_in_responses
    
    async def run_with_tool_live_without_claim_extraction(self, claims):
        queries = await self._query_generation(claims)

        exec_results = []
        for query in queries:
            try:
                exec_results.append(self.tool.run(query['python_snippet']))
            except:
                exec_results.append(None)
        classification_results = self._verification(exec_results)
        return queries, exec_results, classification_results
    
    async def run_with_tool_api_call(self, prompts, responses):
        batch_size = 5
        num_batches = math.ceil(len(prompts) / batch_size)

        self.sample_list = [{"prompt": prompt, "response": response, "category": 'math'} for prompt, response in zip(prompts, responses)]

        for i in range(num_batches):
            print(i)
            batch_start = i * batch_size
            batch_end = min((i + 1) * batch_size, len(responses))

            claims_in_responses, queries_in_responses, exec_results_in_response, verifications_in_responses = await self.run_with_tool_live(self.sample_list[batch_start: batch_end])

            for j, (claims_in_response, queries_in_response, exec_results_in_response, verifications_in_response) in enumerate(zip(claims_in_responses, queries_in_responses, exec_results_in_response, verifications_in_responses)):
                index = batch_start + j

                self.sample_list[index].update({
                    'claims': claims_in_response,
                    'queries': queries_in_response,
                    'execution_results': exec_results_in_response,
                    'claim_level_factuality': verifications_in_response,
                    'response_level_factuality': all([verification if verification != None else True for verification in verifications_in_response])
                })

        return self.sample_list

    async def run_with_tool_dataset(self, annotated_dataset_path: str, with_tool_classified_dataset_path: str, rerun: bool = False, rerun_indices: list = []):
        data_path = annotated_dataset_path if not rerun else with_tool_classified_dataset_path
        with open(data_path, 'r') as f:
            data = [json.loads(line) for line in f]
        self.sample_list = data if rerun else [claim for sample in data for claim in sample['claims']]
        rerun_elements = self.sample_list if not rerun else [self.sample_list[i] for i in rerun_indices]

        batch_size = 10
        num_batches = math.ceil(len(rerun_elements) / batch_size) # 5

        for i in range(num_batches):
            print("test1")
            print(i)
            batch_start = i * batch_size
            batch_end = min((i + 1) * batch_size, len(rerun_elements))
            batch = rerun_elements[batch_start:batch_end]

            queries, exec_results, classification_results = await self.run_with_tool_live_without_claim_extraction(batch)

            for j, (query, exec_result, classification_result) in enumerate(zip(queries, exec_results, classification_results)):
                index = batch_start + j if not rerun else rerun_indices[batch_start + j]
                self.sample_list[index].update({
                    'query': query,
                    'exec_result': exec_result,
                    'with_tool_classification': classification_result,
                })
        
            # save everything after each batch to prevent data loss
            with open(with_tool_classified_dataset_path, 'w') as f:
                for item in self.sample_list:
                    try:
                        json_str = json.dumps(item)
                    except:
                        continue
                    f.write(json_str + '\n')

    async def run_self_check_live(self, fewshot, batch):
        user_prompt_key = 'user_3_shot_CoT' if fewshot else 'user_zero_shot_CoT'
        messages_list = [
            [
                {"role": "system", "content": self.self_check_prompt['system']},
                {"role": "user", "content": self.self_check_prompt[user_prompt_key].format(input_calculation=response['math_calculation'], input_calculated_answer=response['calculated_answer'])},
            ]
            for response in batch
        ]
        return await self.chat.async_run(messages_list, Dict)

    async def run_self_check_dataset(self, annotated_dataset_path: str, self_check_classified_dataset_path: str, fewshot: bool = False, rerun: bool = False, rerun_indices: list = []):
        data_path = annotated_dataset_path if not rerun else self_check_classified_dataset_path
        with open(data_path, 'r') as f:
            data = [json.loads(line) for line in f]
        self.sample_list = data if rerun else [claim for sample in data for claim in sample['claims']]
        rerun_elements = self.sample_list if not rerun else [self.sample_list[i] for i in rerun_indices]

        batch_size = 10
        num_batches = math.ceil(len(rerun_elements) / batch_size)

        for i in range(num_batches):
            print(i)
            batch_start = i * batch_size
            batch_end = min((i + 1) * batch_size, len(rerun_elements))
            batch = rerun_elements[batch_start:batch_end]

            responses = await self.run_self_check_live(fewshot, batch)
            for j, response in enumerate(responses):
                index = batch_start + j if not rerun else rerun_indices[batch_start + j]
                if response is None:
                    self.sample_list[index].update({
                        'self_check_classification': 'None',
                        'self_check_reasoning': 'None'
                    })
                else:
                    self.sample_list[index].update({
                        'self_check_classification': response.get('factuality', 'None'),
                        'self_check_reasoning': response.get('reasoning', 'None')
                    })

            # save everything after each batch to prevent data loss
            with open(self_check_classified_dataset_path, 'w') as f:
                for item in self.sample_list:
                    json_str = json.dumps(item)
                    f.write(json_str + '\n')


from dotagent.memory.base import BaseMemory
from dotagent.memory.read_only import ReadOnlyMemory
from dotagent.memory.in_memory import SimpleMemory
from dotagent.memory.summary import SummaryMemory
from dotagent.memory.buffer_summary import BufferSummaryMemory


__all__=[
    "BaseMemory",
    "ReadOnlyMemory",
    "SimpleMemory",
    "SummaryMemory",
    "BufferSummaryMemory",
]


from typing import List

class SimpleKnowledgeBase:
    def __init__(self, raw_data=None, data_transformer=None, vector_store=None):
        """
        Initialize the knowledge base.

        Args:
            raw_data: The raw data to add to the knowledge base. Default is None.
            data_transformer: An object with a `split_documents` method to apply to the raw data. Default is None.
            vector_store: An object with `add_documents` and `similarity_search` methods to use for storing vectors. Default is None.
        """
        
        self.data_transformer = data_transformer
        self.vector_store = vector_store
        self.references = []
        self.add_data(raw_data)


    def add_data(self, raw_data):
        """
        Add raw data into the knowledge base.

        Args:
            raw_data: The raw data to add.
        """
        # Validate raw data
        if not raw_data:
            raise ValueError("Raw data cannot be empty.")
        
        # fetch and add references
        for data in raw_data:
            self.references.append(data.metadata)
        
        # Split raw data into chunks
        split_data = self.data_transformer.split_documents(raw_data)

        # Add split data to vector store
        try:
            self.vector_store.add_documents(split_data)
        except Exception as e:
            print(f"Failed to add documents: {e}")
            raise

    def retrieve_data(self, query, top_k=1) -> List[str]:
        """
        Retrieve documents from the knowledge base.

        Args:
            query: The query to use for the retrieval.
            top_k: The number of documents to retrieve. Default is 1.
        
        Returns:
            A list of the retrieved documents.
        """
        try:
            results = self.vector_store.similarity_search(query=query, top_k=top_k)
        except Exception as e:
            print(f"Failed to retrieve documents: {e}")
            raise

        # Handle no results case
        if not results:
            return []

        # Extract page content
        docs = [result[0].page_content for result in results]
        return docs


# """Init file."""
# from dotagent.knowledgebase.document_loaders.basereader import BaseReader
# from dotagent.knowledgebase.document_loaders.utils import import_loader
# from dotagent.knowledgebase.document_loaders.airtable.base import AirtableReader
# # from dotagent.knowledgebase.document_loaders.apify.actor.base import ApifyActor
# from dotagent.knowledgebase.document_loaders.apify.dataset.base import ApifyDataset
# from dotagent.knowledgebase.document_loaders.asana.base import AsanaReader
# from dotagent.knowledgebase.document_loaders.azcognitive_search.base import AzCognitiveSearchReader
# # from dotagent.knowledgebase.document_loaders.azstorage_blob.base import AzStorageBlobReader
# from dotagent.knowledgebase.document_loaders.bilibili.base import BilibiliTranscriptReader
# from dotagent.knowledgebase.document_loaders.boarddocs.base import BoardDocsReader
# from dotagent.knowledgebase.document_loaders.chatgpt_plugin.base import ChatGPTRetrievalPluginReader
# from dotagent.knowledgebase.document_loaders.chroma.base import ChromaReader
# from dotagent.knowledgebase.document_loaders.confluence.base import ConfluenceReader
# from dotagent.knowledgebase.document_loaders.couchdb.base import SimpleCouchDBReader
# from dotagent.knowledgebase.document_loaders.dad_jokes.base import DadJokesReader
# # from dotagent.knowledgebase.document_loaders.database.base import DatabaseReader
# from dotagent.knowledgebase.document_loaders.deeplake.base import DeepLakeReader
# from dotagent.knowledgebase.document_loaders.discord.base import DiscordReader
# from dotagent.knowledgebase.document_loaders.docugami.base import DocugamiReader
# from dotagent.knowledgebase.document_loaders.elasticsearch.base import ElasticsearchReader
# from dotagent.knowledgebase.document_loaders.faiss.base import FaissReader
# from dotagent.knowledgebase.document_loaders.feedly_rss.base import FeedlyRssReader
# from dotagent.knowledgebase.document_loaders.feishu_docs.base import FeishuDocsReader
# from dotagent.knowledgebase.document_loaders.file.base import SimpleDirectoryReader
# from dotagent.knowledgebase.document_loaders.file.audio.base import AudioTranscriber
# from dotagent.knowledgebase.document_loaders.file.audio_gladia.base import GladiaAudioTranscriber
# from dotagent.knowledgebase.document_loaders.file.cjk_pdf.base import CJKPDFReader
# from dotagent.knowledgebase.document_loaders.file.deepdoctection.base import DeepDoctectionReader
# from dotagent.knowledgebase.document_loaders.file.docx.base import DocxReader
# from dotagent.knowledgebase.document_loaders.file.epub.base import EpubReader
# from dotagent.knowledgebase.document_loaders.file.flat_pdf.base import FlatPdfReader
# # from dotagent.knowledgebase.document_loaders.file.image.base import ImageReader
# # from dotagent.knowledgebase.document_loaders.file.image_blip.base import ImageCaptionReader
# # from dotagent.knowledgebase.document_loaders.file.image_blip2.base import ImageVisionLLMReader
# # from dotagent.knowledgebase.document_loaders.file.image_deplot.base import ImageTabularChartReader
# from dotagent.knowledgebase.document_loaders.file.ipynb.base import IPYNBReader
# from dotagent.knowledgebase.document_loaders.file.json.base import JSONReader
# from dotagent.knowledgebase.document_loaders.file.markdown.base import MarkdownReader
# from dotagent.knowledgebase.document_loaders.file.mbox.base import MboxReader
# from dotagent.knowledgebase.document_loaders.file.paged_csv.base import PagedCSVReader
# from dotagent.knowledgebase.document_loaders.file.pandas_csv.base import PandasCSVReader
# from dotagent.knowledgebase.document_loaders.file.pandas_excel.base import PandasExcelReader
# from dotagent.knowledgebase.document_loaders.file.pdf.base import PDFReader
# from dotagent.knowledgebase.document_loaders.file.pdf_miner.base import PDFMinerReader
# from dotagent.knowledgebase.document_loaders.file.pptx.base import PptxReader
# from dotagent.knowledgebase.document_loaders.file.pymu_pdf.base import PyMuPDFReader
# from dotagent.knowledgebase.document_loaders.file.rdf.base import RDFReader
# from dotagent.knowledgebase.document_loaders.file.simple_csv.base import SimpleCSVReader
# from dotagent.knowledgebase.document_loaders.file.unstructured.base import UnstructuredReader
# from dotagent.knowledgebase.document_loaders.firebase_realtimedb.base import FirebaseRealtimeDatabaseReader
# from dotagent.knowledgebase.document_loaders.firestore.base import FirestoreReader
# # from dotagent.knowledgebase.document_loaders.github_repo.base import GithubRepositoryReader
# from dotagent.knowledgebase.document_loaders.github_repo_issues.base import GitHubRepositoryIssuesReader
# from dotagent.knowledgebase.document_loaders.gmail.base import GmailReader
# from dotagent.knowledgebase.document_loaders.google_calendar.base import GoogleCalendarReader
# from dotagent.knowledgebase.document_loaders.google_docs.base import GoogleDocsReader
# # from dotagent.knowledgebase.document_loaders.google_drive.base import GoogleDriveReader
# from dotagent.knowledgebase.document_loaders.google_keep.base import GoogleKeepReader
# from dotagent.knowledgebase.document_loaders.google_sheets.base import GoogleSheetsReader
# from dotagent.knowledgebase.document_loaders.gpt_repo.base import GPTRepoReader
# from dotagent.knowledgebase.document_loaders.graphdb_cypher.base import GraphDBCypherReader
# from dotagent.knowledgebase.document_loaders.graphql.base import GraphQLReader
# from dotagent.knowledgebase.document_loaders.hatena_blog.base import HatenaBlogReader
# from dotagent.knowledgebase.document_loaders.hubspot.base import HubspotReader
# from dotagent.knowledgebase.document_loaders.huggingface.fs.base import HuggingFaceFSReader
# from dotagent.knowledgebase.document_loaders.intercom.base import IntercomReader
# from dotagent.knowledgebase.document_loaders.jira.base import JiraReader
# # from dotagent.knowledgebase.document_loaders.joplin.base import JoplinReader
# from dotagent.knowledgebase.document_loaders.jsondata.base import JSONDataReader
# from dotagent.knowledgebase.document_loaders.kaltura.esearch.base import KalturaESearchReader
# from dotagent.knowledgebase.document_loaders.kibela.base import  KibelaReader
# # from dotagent.knowledgebase.document_loaders.make_com.base import MakeWrapper
# from dotagent.knowledgebase.document_loaders.mangoapps_guides.base import MangoppsGuidesReader
# from dotagent.knowledgebase.document_loaders.maps.base import OpenMap
# from dotagent.knowledgebase.document_loaders.memos.base import MemosReader
# from dotagent.knowledgebase.document_loaders.metal.base import MetalReader
# from dotagent.knowledgebase.document_loaders.milvus.base import MilvusReader
# from dotagent.knowledgebase.document_loaders.mondaydotcom.base import MondayReader
# from dotagent.knowledgebase.document_loaders.mongo.base import SimpleMongoReader
# from dotagent.knowledgebase.document_loaders.notion.base import NotionPageReader
# # from dotagent.knowledgebase.document_loaders.obsidian.base import ObsidianReader
# # from dotagent.knowledgebase.document_loaders.opendal_reader.base import OpendalReader
# # from dotagent.knowledgebase.document_loaders.opendal_reader.azblob.base import OpendalAzblobReader
# # from dotagent.knowledgebase.document_loaders.opendal_reader.gcs.base import OpendalGcsReader
# # from dotagent.knowledgebase.document_loaders.opendal_reader.s3.base import OpendalS3Reader
# from dotagent.knowledgebase.document_loaders.outlook_localcalendar.base import OutlookLocalCalendarReader
# # from dotagent.knowledgebase.document_loaders.pandas_ai.base import PandasAIReader
# # from dotagent.knowledgebase.document_loaders.papers.arxiv.base import ArxivReader
# from dotagent.knowledgebase.document_loaders.papers.pubmed.base import PubmedReader
# from dotagent.knowledgebase.document_loaders.pinecone.base import PineconeReader
# from dotagent.knowledgebase.document_loaders.qdrant.base import QdrantReader
# from dotagent.knowledgebase.document_loaders.readwise.base import ReadwiseReader
# from dotagent.knowledgebase.document_loaders.reddit.base import RedditReader
# # from dotagent.knowledgebase.document_loaders.remote.base import RemoteReader
# # from dotagent.knowledgebase.document_loaders.remote_depth.base import RemoteDepthReader
# # from dotagent.knowledgebase.document_loaders.s3.base import S3Reader
# # from dotagent.knowledgebase.document_loaders.singlestore.base import SingleStoreReader
# from dotagent.knowledgebase.document_loaders.slack.base import SlackReader
# from dotagent.knowledgebase.document_loaders.snscrape_twitter.base import SnscrapeTwitterReader
# from dotagent.knowledgebase.document_loaders.spotify.base import SpotifyReader
# from dotagent.knowledgebase.document_loaders.stackoverflow.base import StackoverflowReader
# from dotagent.knowledgebase.document_loaders.steamship.base import SteamshipFileReader
# from dotagent.knowledgebase.document_loaders.string_iterable.base import StringIterableReader
# from dotagent.knowledgebase.document_loaders.trello.base import TrelloReader
# from dotagent.knowledgebase.document_loaders.twitter.base import TwitterTweetReader
# from dotagent.knowledgebase.document_loaders.weather.base import WeatherReader
# from dotagent.knowledgebase.document_loaders.weaviate.base import WeaviateReader
# from dotagent.knowledgebase.document_loaders.web.async_web.base import AsyncWebPageReader
# from dotagent.knowledgebase.document_loaders.web.beautiful_soup_web.base import BeautifulSoupWebReader
# from dotagent.knowledgebase.document_loaders.web.knowledge_base.base import KnowledgeBaseWebReader
# # from dotagent.knowledgebase.document_loaders.web.readability_web.base import ReadabilityWebPageReader
# from dotagent.knowledgebase.document_loaders.web.rss.base import RssReader
# from dotagent.knowledgebase.document_loaders.web.simple_web.base import SimpleWebPageReader
# # from dotagent.knowledgebase.document_loaders.web.sitemap.base import SitemapReader
# from dotagent.knowledgebase.document_loaders.web.trafilatura_web.base import TrafilaturaWebReader
# from dotagent.knowledgebase.document_loaders.web.unstructured_web.base import UnstructuredURLLoader
# from dotagent.knowledgebase.document_loaders.whatsapp.base import WhatsappChatLoader
# from dotagent.knowledgebase.document_loaders.wikipedia.base import WikipediaReader
# from dotagent.knowledgebase.document_loaders.wordlift.base import WordLiftLoader
# from dotagent.knowledgebase.document_loaders.wordpress.base import WordpressReader
# from dotagent.knowledgebase.document_loaders.youtube_transcript.base import YoutubeTranscriptReader
# from dotagent.knowledgebase.document_loaders.zendesk.base import ZendeskReader
# from dotagent.knowledgebase.document_loaders.zulip.base import ZulipReader



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import json

class Tools:
    @staticmethod
    def load_jsonl(file_path):
        json_objects = []
        with open(file_path, 'r', encoding='utf8') as f:
            for line in f:
                json_objects.append(json.loads(line.strip()))
        return json_objects
    
    @staticmethod
    def load_tasks(task_path):
        result = dict()
        lines = Tools.load_jsonl(task_path)
        for line in lines:
            result[line['task_id']] = line
        return result
        
    @staticmethod
    def write_file(path, content):
        with open(path, 'w', encoding='utf8') as f:
            f.write(content)


import time
import json
import math
import os
import yaml
from typing import Dict, List

from dotagent.fact_check.scientific.tool import google_scholar
from dotagent.fact_check.utils.base.pipeline import pipeline

class scientific_pipeline(pipeline):
    def __init__(self, foundation_model):
        super().__init__('scientific', foundation_model)

        self.tool = google_scholar()

        with open(os.path.join(self.prompts_path, "claim_extraction.yaml"), 'r') as file:
            data = yaml.load(file, Loader=yaml.FullLoader)
        self.claim_prompt = data['scientific']

        with open(os.path.join(self.prompts_path, 'agreement_verification.yaml'), 'r') as file:
            data = yaml.load(file, Loader=yaml.FullLoader)
        self.verification_prompt = data['scientific']

    async def _claim_extraction(self, responses):
        messages_list = [
            [
                {"role": "system", "content": self.claim_prompt['system']},
                {"role": "user", "content": self.claim_prompt['user'].format(input=response)},
            ]
            for response in responses
        ]
        return await self.chat.async_run(messages_list, List)
    
    async def _check_authors(self, authors):
        messages_list = [
            [
                {"role": "system", "content": self.verification_prompt['system']},
                {"role": "user", "content": self.verification_prompt['user'].format(string1=claim_author, list2=real_author)},
            ]
            for claim_author, real_author in authors
        ]

        return await self.chat.async_run(messages_list, Dict)

    async def _verification(self, claims, responses):
        authors = [(claim['paper_author(s)'], response['author']) for claim, response in zip(claims, responses)]
        check_authors_results = await self._check_authors(authors)
        final_responses = []
        for i, (claim, response) in enumerate(zip(claims, responses)):
            final_response = {
                'generated_paper_title': claim['paper_title'],
                'generated_paper_author(s)': claim['paper_author(s)'],
                'generated_paper_pub_year': claim['paper_pub_year'],
                'actual_paper_title': response['title'],
                'actual_paper_author(s)': response['author'],
                'actual_paper_pub_year': response['pub_year'],
            }

            errors = []
            if final_response['generated_paper_title'].lower() != final_response['actual_paper_title'].lower() and final_response['generated_paper_title'].lower() not in final_response['actual_paper_title'].lower() and final_response['actual_paper_title'].lower() not in final_response['generated_paper_title'].lower() :
                errors.append('wrong_paper_title')
            if check_authors_results[i]['factuality'] == False:
                errors.append('wrong_paper_author(s)')
            if final_response['generated_paper_pub_year'] != final_response['actual_paper_pub_year']:
                errors.append('wrong_paper_pub_year')

            final_response['error'] = errors
            final_response['factuality'] = len(errors) == 0

            final_responses.append(final_response)

        return final_responses

    async def run_with_tool_live(self, samples):
        claims_in_responses = await self._claim_extraction(samples)
        queries_in_responses = []
        evidences_in_responses = []
        verifications_in_responses = []
        for claims_in_response in claims_in_responses:
            queries = [claim['paper_title'] for claim in claims_in_response]
            queries_in_responses.append(queries)
            evidences = [self.tool.run(paper_title) for paper_title in queries]
            evidences_in_responses.append(evidences)
            verifications = await self._verification(claims_in_response, evidences)
            verifications_in_responses.append(verifications)

        return claims_in_responses, queries_in_responses, evidences_in_responses, verifications_in_responses
        
    async def run_with_tool_live_without_claim_extraction(self, claims):
        # claims = [{"paper_title": "A Survey of Modern Authorship Attribution Methods", "paper_author(s)": "Stamatatos, Efstathios", "paper_pub_year": "2013"}, {"paper_title": "BERT", "paper_author(s)": "John Smith", "paper_pub_year": "2020"}]
        papers_titles = [claim['paper_title'] for claim in claims]
        responses = [self.tool.run(paper_title) for paper_title in papers_titles]
        final_response = await self._verification(claims, responses)
        return final_response

    async def run_with_tool_api_call(self, prompts, responses):
        batch_size = 5
        num_batches = math.ceil(len(prompts) / batch_size)

        self.sample_list = [{"prompt": prompt, "response": response, "category": 'scientific'} for prompt, response in zip(prompts, responses)]

        for i in range(num_batches):
            print(i)
            batch_start = i * batch_size
            batch_end = min((i + 1) * batch_size, len(responses))

            claims_in_responses, queries_in_responses, evidences_in_responses, verifications_in_responses = await self.run_with_tool_live(responses[batch_start:batch_end])

            for j, (claims_in_response, queries_in_response, evidences_in_response, verifications_in_response) in enumerate(zip(claims_in_responses, queries_in_responses, evidences_in_responses, verifications_in_responses)):
                index = batch_start + j

                self.sample_list[index].update({
                    'claims': claims_in_response,
                    'queries': queries_in_response,
                    'evidences': evidences_in_response,
                    'claim_level_factuality': verifications_in_response,
                    'response_level_factuality': all([verification['factuality'] if verification != None else True for verification in verifications_in_response])
                })
        return self.sample_list

    async def run_with_tool_dataset(self, annotated_dataset_path: str, with_tool_classified_dataset_path: str, rerun: bool = False, rerun_indices: list = []):
        # Example of a line: 
        # {"paper_title": "A Survey of Modern Authorship Attribution Methods", "paper_author(s)": "Stamatatos, Efstathios", "paper_pub_year": "2013", "label": True / False}
        if rerun == False:
            with open(annotated_dataset_path, 'r') as f:
                data = [json.loads(line) for line in f]
            self.sample_list = [claim for sample in data for claim in sample['claims']]
            rerun_elements = self.sample_list
        else:
            with open(with_tool_classified_dataset_path, 'r') as f:
                data = [json.loads(line) for line in f]
            self.sample_list = data
            rerun_elements = [self.sample_list[i] for i in rerun_indices]

        batch_size = 5
        num_batches = math.ceil(len(rerun_elements) / batch_size) # 5

        for i in range(num_batches):
            print(i)
            batch_start = i * batch_size
            batch_end = (i + 1) * batch_size if (i + 1) * batch_size < len(rerun_elements) else len(rerun_elements)

            responses = await self.run_with_tool_live_without_claim_extraction(rerun_elements[batch_start:batch_end])
            for j, response in enumerate(responses):
                index = batch_start + j if rerun == False else rerun_indices[batch_start + j]
                if response == None:
                    self.sample_list[index]['with_tool_classification'] = 'None'
                    self.sample_list[index]['error'] = 'None'
                else:
                    self.sample_list[index]['with_tool_classification'] = response.get('factuality', 'None')
                    self.sample_list[index]['error'] = response.get('error', 'None')
        
            # save everything after each batch to prevent data loss
            with open(with_tool_classified_dataset_path, 'w') as f:
                for item in self.sample_list:
                    json_str = json.dumps(item)
                    f.write(json_str + '\n')

    async def run_self_check_live(self, fewshot, batch):
        user_prompt_key = 'user_3_shot_CoT' if fewshot else 'user_zero_shot_CoT'
        messages_list = [
            [
                {"role": "system", "content": self.self_check_prompt['system']},
                {"role": "user", "content": self.self_check_prompt[user_prompt_key].format(scientific_literature=response)}
            ]
            for response in batch
        ]
        return await self.chat.async_run(messages_list, Dict)

    async def run_self_check_dataset(self, annotated_dataset_path: str, self_check_classified_dataset_path: str, fewshot: bool = False, rerun: bool = False, rerun_indices: list = []):
        # Example of a line: 
        # {"paper_title": "A Survey of Modern Authorship Attribution Methods", "paper_author(s)": "Stamatatos, Efstathios", "paper_pub_year": "2013", "annotation": True / False}
        data_path = annotated_dataset_path if not rerun else self_check_classified_dataset_path
        with open(data_path, 'r') as f:
            data = [json.loads(line) for line in f]
        self.sample_list = data if rerun else [claim for sample in data for claim in sample['claims']]
        rerun_elements = self.sample_list if not rerun else [self.sample_list[i] for i in rerun_indices]

        batch_size = 5
        num_batches = math.ceil(len(rerun_elements) / batch_size)

        for i in range(num_batches):
            print(i)
            batch_start = i * batch_size
            batch_end = (i + 1) * batch_size
            batch = rerun_elements[batch_start:batch_end]
            batch = [{k:v for k,v in d.items() if k != "label"} for d in batch]

            responses = await self.run_self_check_live(fewshot, batch)
            for j, response in enumerate(responses):
                index = batch_start + j if rerun == False else rerun_indices[batch_start + j]
                if response == None:
                    self.sample_list[index]['self_check_classification'] = 'None'
                    self.sample_list[index]['self_check_reasoning'] = 'None'
                else:
                    self.sample_list[index]['self_check_classification'] = response.get('factuality', 'None')
                    self.sample_list[index]['self_check_reasoning'] = response.get('reasoning', 'None')
        
            # save everything after each batch to prevent data loss
            with open(self_check_classified_dataset_path, 'w') as f:
                for item in self.sample_list:
                    json_str = json.dumps(item)
                    f.write(json_str + '\n')


""" Utility functions for API requests """

import json
from io import BytesIO
from typing import Optional
import requests
from aiohttp import ClientResponse, ClientSession, FormData
from aiohttp.payload import BytesIOPayload
from dotagent.config import settings


class AgentBoxError(Exception):
    """
    Represents an api error returned from the AgentBox API.
    """

    def __init__(
        self,
        http_status: int = 0,
        json_body: dict = {},
        headers: dict = {},
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.http_status = http_status
        self.json_body = json_body
        self.headers = headers

    def __str__(self):
        return f"{self.http_status}: {self.json_body}"

    def __repr__(self):
        return f"<AgentBoxError {self.http_status}: {self.json_body}>"

def build_agentbox_request_data(
    method: str,
    endpoint: str,
    body: Optional[dict] = None,
    files: Optional[dict] = None,
) -> dict:
    """
    Builds a request data dictionary for the requests library.
    """
    return {
        "method": method,
        "url": settings.AGENTBOX_BASE_URL + endpoint,
        "headers": {
            "Authorization": f"Bearer {settings.AGENTBOX_API_KEY}",
        },
        "json": body,
        "files": files,
    }


def handle_agentbox_response(response: requests.Response):
    """
    Handles a response from the requests library.
    """
    handlers = {
        "application/json": lambda r: json.loads(r.content.decode()),
        "application/octet-stream": lambda r: {
            "content": BytesIO(r.content).read(),
            "name": r.headers["Content-Disposition"].split("=")[1],
        }
        # Add other content type handlers here
    }
    handler = handlers.get(
        response.headers["Content-Type"].split(";")[0], lambda r: r.content.decode()
    )
    if response.status_code != 200:
        raise AgentBoxError(
            http_status=response.status_code,
            json_body=response.json(),
            headers=response.headers.__dict__,
        )
    return handler(response)


async def handle_response_async(response: ClientResponse) -> dict:
    """
    Handles a response from the aiohttp library.
    """

    async def json_handler(r: ClientResponse) -> dict:
        return json.loads(await r.text())

    async def file_handler(r: ClientResponse) -> dict:
        return {
            "content": await r.read(),
            "name": r.headers["Content-Disposition"].split("=")[1],
        }

    async def default_handler(r: ClientResponse) -> dict:
        return {"content": await r.text()}

    handlers = {
        "application/json": json_handler,
        "application/octet-stream": file_handler,
        # Add other content type handlers here
    }
    handler = handlers.get(
        response.headers["Content-Type"].split(";")[0], default_handler
    )
    if response.status != 200:
        raise AgentBoxError(
            http_status=response.status,
            json_body=await response.json(),
            headers=response.headers.__dict__,
        )
    return await handler(response)


def base_request(
    method: str,
    endpoint: str,
    body: Optional[dict] = None,
    files: Optional[dict] = None,
) -> dict:
    """
    Makes a request to the AgentBox API.
    """
    request_data = build_request_data(method, endpoint, body, files)
    response = requests.request(**request_data, timeout=90)
    return handle_response(response)


async def abase_request(
    session: ClientSession,
    method: str,
    endpoint: str,
    body: Optional[dict] = None,
    files: Optional[dict] = None,
) -> dict:
    """
    Makes an asynchronous request to the AgentBox API.
    """
    request_data = build_request_data(method, endpoint, body, files)
    if files is not None:
        data = FormData()
        for key, file_tuple in files.items():
            filename, fileobject = file_tuple[
                :2
            ]  # Get the filename and fileobject from the tuple
            payload = BytesIOPayload(BytesIO(fileobject))
            data.add_field(
                key, payload, filename=filename
            )  # Use the filename from the tuple
        request_data.pop("files")
        request_data.pop("json")
        request_data["data"] = data
        response = await session.request(**request_data)
    else:
        request_data.pop("files")
        response = await session.request(**request_data)
    return await handle_response_async(response)


def set_agentbox_api_key(api_key: str) -> None:
    """
    Manually set the AGENTBOX_API_KEY.
    """
    settings.AGENTBOX_API_KEY = api_key




from typing import List, Any
from pydantic import BaseModel

from dotagent.memory import BaseMemory


class SimpleMemory(BaseMemory, BaseModel):
    def add_memory(self, prompt: str, llm_response: Any) -> None:
        """Add a self-created message to the store"""
        unique = True
        for conversation in self.messages:
            if conversation['prompt'] == prompt and conversation['llm_response']  == llm_response:
                unique = False
        if unique:
            self.messages.append({'prompt': prompt, 'llm_response': llm_response})

    def get_memory(self, **kwargs) -> str:
        """Retrieve entire memory from the store."""
        if len(self.messages) != 0:
            messages_to_text = "Current conversation:\n"
            for conversation in self.messages:
                messages_to_text = messages_to_text + "Human: " + conversation['prompt'] + "\n" + "AI: " + conversation['llm_response'] + "\n"
        else:
            messages_to_text= ""
        return messages_to_text

    def remove_memory(self, prompt: str) -> None:
        """Remove a memory from the store."""
        for conversation in self.messages:
            if conversation['prompt'] == prompt:
                self.messages.remove(conversation)
                break

    def clear(self) -> None:
        """Clear all memories."""
        self.messages.clear()


"""Zapier tool spec."""

from dotagent.tools.basetool import BaseTool
from typing import Optional, Dict, Type
import requests
import json
from pydantic import BaseModel, Field

ACTION_URL_TMPL = "https://nla.zapier.com/api/v1/dynamic/exposed/{action_id}/execute/"
class NaturalLanguageSchema(BaseModel):
    id : str = Field(
        ...,
        description="Info about the Parameter"
    )
    kwargs = Field(
        ...,
        description="Info about the Paramter"
    )

class BaseZapier(BaseTool):
    """Zapier tool spec."""
    def __init__(self, api_key: Optional[str] = None, oauth_access_token: Optional[str] = None) -> None:
        """Initialize with parameters."""
        if api_key:
            self._headers = {
                "x-api-key": api_key
                }
        elif oauth_access_token:
            self._headers = {
                "Authorization": f"Bearer {oauth_access_token}"
            }
        else:
            raise ValueError("Must provide either api_key or oauth_access_token")

        # Get the exposed actions from Zapier
        actions = json.loads(self.list_actions())
        if "results" not in actions:
            raise ValueError("No Zapier actions exposed, visit https://nla.zapier.com/dev/actions/ to expose actions.")
        results = actions["results"]

        # Register the actions as Tools
        for action in results:
            params = action['params']

            def function_action(id=action['id'], **kwargs):
                return self.natural_language_query(id, **kwargs)

            action_name = action['description'].split(': ')[1].replace(' ', '_')
            function_action.__name__ = action_name
            function_action.__doc__ = f"""
                This is a Zapier Natural Language Action function wrapper.

                The 'instructions' key is REQUIRED for all function calls.
                The instructions key is a natural language string describing the action to be taken
                The following are all of the valid arguments you can provide: {params}

                Ignore the id field, it is provided for you.
                If the returned error field is not null, interpret the error and try to fix it. Otherwise, inform the user of how they might fix it.
            """
            setattr(
                self,
                action_name,
                function_action
            )
            self.spec_functions.append(action_name)


class ListAction(BaseZapier):
    name: str = "ListAction"
    description: str = "Description of the method"
    def list_actions(self):
        response = requests.get(
            "https://nla.zapier.com/api/v1/dynamic/exposed/",
            headers=self._headers
        )
        return response.text

    def run(self):
        try:
            self.list_actions()
        except Exception as e:
            return e

class NaturalLanguageQuery(BaseZapier):
    name: str = "NaturalLanguageQuery"
    description: str = "Description of the method"
    args_schema: Type[NaturalLanguageSchema] = NaturalLanguageSchema
    def natural_language_query(self, id: str, **kwargs):
        response = requests.post(
            ACTION_URL_TMPL.format(action_id=id),
            headers=self._headers,
            data=json.dumps(kwargs)
        )
        return response.text
    
    def run(self, id: str, **kwargs):
        try:
            self.natural_language_query(self, id=id, kwargs=kwargs)
        except Exception as e:
            return e


from abc import ABC, abstractmethod
from logging import Logger
import openai

class LLMFinetune(ABC):
    def __init__(self, logger: Logger, openai_key: str):
        self.logger = logger
        openai.api_key = openai_key

    @abstractmethod
    def transform_data(self, train_csv_file: str, val_csv_file: str , train_output_file: str, val_output_file: str) -> str:
        pass

    @abstractmethod
    def finetune(self, **kwargs):
        pass



from abc import ABC, abstractmethod


class SpeechToText(ABC):
    @abstractmethod
    def transcribe(self, audio_bytes, platform='web', prompt='', language='en-US') -> str:
        # platform: 'web' | 'mobile' | 'terminal'
        pass



""" Abstract Base Class for Isolated Execution Environments (AgentBox's) """

from abc import ABC, abstractmethod
from datetime import datetime
from os import PathLike
from queue import Queue
import threading
from typing import List, Optional
from uuid import UUID
from dotagent.schema import AgentBoxFile, AgentBoxOutput, AgentBoxStatus


class BaseBox(ABC):
    """AgentBox Abstract Base Class"""

    def __init__(self, session_id: Optional[UUID] = None) -> None:
        """Initialize the AgentBox instance"""
        self.session_id = session_id
        self.last_interaction = datetime.now()
        self.message_queue = Queue()
        self.message_thread = threading.Thread(target=self._message_handler)
        self.message_thread.start()

    def _update(self) -> None:
        """Update last interaction time"""
        self.last_interaction = datetime.now()

    @abstractmethod
    def start(self) -> AgentBoxStatus:
        """Startup the AgentBox instance"""

    @abstractmethod
    async def astart(self) -> AgentBoxStatus:
        """Async Startup the AgentBox instance"""
    
    @abstractmethod
    def _message_handler(self):
        """Abstract method for handling messages from the queue."""
        pass

    @abstractmethod
    def _send_message_to_kernel(self, message):
        """Abstract method for sending a message to the kernel."""
        pass

    @abstractmethod
    def status(self) -> AgentBoxStatus:
        """Get the current status of the AgentBox instance"""

    @abstractmethod
    async def astatus(self) -> AgentBoxStatus:
        """Async Get the current status of the AgentBox instance"""

    @abstractmethod
    def run(
        self, code: Optional[str] = None, file_path: Optional[PathLike] = None
    ) -> AgentBoxOutput:
        """Execute python code inside the AgentBox instance"""

    @abstractmethod
    async def arun(
        self, code: str, file_path: Optional[PathLike] = None
    ) -> AgentBoxOutput:
        """Async Execute python code inside the AgentBox instance"""

    @abstractmethod
    def upload(self, file_name: str, content: bytes) -> AgentBoxStatus:
        """Upload a file as bytes to the AgentBox instance"""

    @abstractmethod
    async def aupload(self, file_name: str, content: bytes) -> AgentBoxStatus:
        """Async Upload a file as bytes to the AgentBox instance"""

    @abstractmethod
    def download(self, file_name: str) -> AgentBoxFile:
        """Download a file as AgentBoxFile schema"""

    @abstractmethod
    async def adownload(self, file_name: str) -> AgentBoxFile:
        """Async Download a file as AgentBoxFile schema"""

    @abstractmethod
    def install(self, package_name: str) -> AgentBoxStatus:
        """Install a python package to the venv"""

    @abstractmethod
    async def ainstall(self, package_name: str) -> AgentBoxStatus:
        """Async Install a python package to the venv"""

    @abstractmethod
    def list_files(self) -> List[AgentBoxFile]:
        """List all available files inside the AgentBox instance"""

    @abstractmethod
    async def alist_files(self) -> List[AgentBoxFile]:
        """Async List all available files inside the AgentBox instance"""

    @abstractmethod
    def restart(self) -> AgentBoxStatus:
        """Restart the jupyter kernel inside the AgentBox instance"""

    @abstractmethod
    async def arestart(self) -> AgentBoxStatus:
        """Async Restart the jupyter kernel inside the AgentBox instance"""

    @abstractmethod
    def stop(self) -> AgentBoxStatus:
        """Terminate the AgentBox instance"""
        self.message_queue.put("STOP")
        self.message_thread.join()
        return self._stop()
    
    @abstractmethod
    def _stop(self) -> AgentBoxStatus:
        """Abstract method for stopping the AgentBox instance"""
        pass

    @abstractmethod
    async def astop(self) -> AgentBoxStatus:
        """Async Terminate the AgentBox instance"""
        # Stop the message handling thread
        self.message_queue.put("STOP")
        await self.message_thread.join() 
        return await self._astop()

    @abstractmethod
    async def _astop(self) -> AgentBoxStatus:
        """Abstract method for asynchronously stopping the AgentBox instance"""
        pass

    def __enter__(self) -> "BaseBox":
        self.start()
        return self

    async def __aenter__(self) -> "BaseBox":
        await self.astart()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.stop()

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:
        await self.astop()

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.session_id}>"

    def __str__(self) -> str:
        return self.__repr__()





"""
AgentBox is the simplest excution infrastructure for your LLM Apps and Services.
"""

from basebox import BaseBox
from agentbox import AgentBox
from tinybox import TinyBox
from dotagent.config import settings
from . _utils import set_api_key

__all__ = [
    "BaseBox",
    "AgentBox",
    "TinyBox",
    "set_api_key",
    "settings",
]



from dotagent.schema import Document
from dotagent.tools.toolkits.slack_toolkit.slack_tool.utils import SlackReader
from dotagent.tools.basetool import BaseTool
from typing import Optional, List, Type
from ssl import SSLContext
from pydantic import BaseModel, Field
from datetime import datetime
import pydantic
import logging

logger = logging.getLogger(__name__)

class SlackBase(BaseTool):

    """Slack toolkit."""

    reader: Optional[SlackReader] = Field(None)
    slack_token: Optional[str] = Field(None)
    ssl: Optional[SSLContext] = Field(None)
    earliest_date: Optional[datetime] = Field(None)
    latest_date: Optional[datetime] = Field(None)

    class Config:
        arbitrary_types_allowed = True

    @pydantic.validator('reader', pre=True, always=True)
    def set_reader(cls, v, values):
        return SlackReader(
            slack_token=values.get('slack_token'),
            ssl=values.get('ssl'),
            earliest_date=values.get('earliest_date'),
            latest_date=values.get('latest_date'),
        )

class LoadDataArgsSchema(BaseModel):
    channel_ids: List[str] = Field(
        ...,
        description="List of IDs for the Slack channels from which to load data",
    )
    reverse_chronological: bool = Field(
        ...,
        description="Signifies whether the loaded data should be ordered in reverse chronological order. By default, it's set to True",
    )

class LoadData(SlackBase):
    name: str = "load_data"
    description: str = "Load data from the input directory."
    args_schema : Type[LoadDataArgsSchema] = LoadDataArgsSchema

    def load_data(
            self,
            channel_ids: List[str],
            reverse_chronological: bool = True,
        ) -> List[Document]:
            """Load data from the input directory."""
            return self.reader.load_data(
                channel_ids=channel_ids,
                reverse_chronological=reverse_chronological,
            )
    def run(
            self,
            channel_ids: List[str],
            reverse_chronological: bool = True,
        )-> str:
         """Run the tool."""
         try:
            return self.load_data(self, channel_ids=channel_ids, reverse_chronological=reverse_chronological)
         except Exception as e:
              raise Exception(f"An error occurred: {e}")
        

class FetchChannel(SlackBase):
    name: str = "fetch_channel"
    description: str = "Fetch a list of relevant channels"
    
    def fetch_channels(
            self,
        ) -> List[str]:
            """Fetch a list of relevant channels."""
            slack_client = self.reader.client
            try:
                msg_result = slack_client.conversations_list()
                logger.info(msg_result)
            except Exception as e:
                logger.error(e)
                raise e

            return msg_result["channels"]
    def _run(
            self,
        )-> str:
         """Run the tool."""
         return self.fetch_channels()

class SendMessageArgsSchema(BaseModel):
    channel_id: str = Field(
        ...,
        description="ID of the channel to send message to",
    )
    message: str = Field(
        ...,
        description="Content of the message",
    )
class SendMessage(SlackBase):
    name: str = "send_message"
    description: str = "Send a message to a channel given the channel ID."
    args_schema : Type[SendMessageArgsSchema] = SendMessageArgsSchema
    def send_message(
            self,
            channel_id: str,
            message: str,
        ) -> None:
            """Send a message to a channel given the channel ID."""
            slack_client = self.reader.client
            try:
                msg_result = slack_client.chat_postMessage(
                    channel=channel_id,
                    text=message,
                )
                logger.info(msg_result)
            except Exception as e:
                logger.error(e)
                raise e
    
    def _run(
            self,
            channel_id: str,
            message: str,
        )-> str:
         """Run the tool."""
         try:
            self.send_message(channel_id=channel_id, message=message)
            return "Message Sent"
         except Exception as e:
            raise Exception(f"An error occurred: {e}")
        


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

from typing import Optional, Dict
import contextlib
import faulthandler
import io
import os
import multiprocessing
import platform
import signal
import tempfile

def test_case_against_solution(prompt: str, completion: str, test: str, timeout: float):
    """
    Get the output of a testcases against a solution
    """

    def unsafe_execute():

        with create_tempdir():

            # These system calls are needed when cleaning up tempdir.
            import os
            import shutil
            rmtree = shutil.rmtree
            rmdir = os.rmdir
            chdir = os.chdir

            # Disable functionalities that can make destructive changes to the test.
            reliability_guard()

            # Construct the check program and run it. VERY IMPORTANT!!!
            import_packages_end_index = prompt.find("def")

            # some HumanEval prompts contain helper function that has
            # to be included. Patching.
            helper_func = ""
            helper_funcs_to_be_included = \
                ["\n\ndef is_palindrome(string: str) -> bool:\n"
                 "    \"\"\" Test if given string is a palindrome \"\"\"\n"
                 "    return string == string[::-1]\n\n\n",
                 "import math\n\n\ndef poly(xs: list, x: float):\n"
                 "    \"\"\"\n    Evaluates polynomial with"
                 " coefficients xs at point x.\n"
                 "    return xs[0] + xs[1] * x + xs[1] * x^2 + .... xs[n] * x^n\n"
                 "    \"\"\"\n    return sum([coeff * math.pow(x, i) for i,"
                 " coeff in enumerate(xs)])\n\n\n", "\n\ndef encode_cyclic(s: str):\n"
                                                    "    \"\"\"\n    returns encoded"
                                                    " string by cycling groups of three"
                                                    " characters.\n    \"\"\"\n"
                                                    "    # split string to groups."
                                                    " Each of length 3.\n"
                                                    "    groups"
                                                    " = [s[(3 * i):min((3 * i + 3),"
                                                    " len(s))] for"
                                                    " i in range((len(s) + 2) // 3)]\n"
                                                    "    # cycle elements in each"
                                                    " group. Unless group has fewer"
                                                    " elements than 3.\n    "
                                                    "groups = [(group[1:] + group[0])"
                                                    " if len(group) == 3 "
                                                    "else group for group in groups]\n"
                                                    "    "
                                                    "return \"\".join(groups)\n\n\n",
                 "\n\ndef encode_shift(s: str):\n    \"\"\"\n"
                 "    returns encoded string by shifting every"
                 " character by 5 in the alphabet.\n    \"\"\"\n"
                 "    return \"\".join([chr(((ord(ch)"
                 " + 5 - ord(\"a\")) % 26) + ord(\"a\")) for ch in s])\n\n\n"]
            for helper in helper_funcs_to_be_included:
                if helper in prompt:
                    helper_func = helper

            check_program = (
                (prompt[:import_packages_end_index] if import_packages_end_index != -1 else "") + helper_func + completion + "\n" + f'result = {test}'
            )

            try:
                exec_globals = {}
                with swallow_io():
                    with time_limit(timeout):
                        exec(check_program, exec_globals)
                        result.append(exec_globals['result'])
            except TimeoutException:
                result.append("FAILURE: timed out")
            except BaseException as e:
                error_type = type(e).__name__
                result.append(f"FAILURE: {error_type}: {e}")

            # Needed for cleaning up.
            shutil.rmtree = rmtree
            os.rmdir = rmdir
            os.chdir = chdir

    manager = multiprocessing.Manager()
    result = manager.list()

    p = multiprocessing.Process(target=unsafe_execute)
    p.start()
    p.join(timeout=timeout+1)
    if p.is_alive():
        p.kill()

    if not result:
        result.append("timed out")

    return result[0]

@contextlib.contextmanager
def time_limit(seconds: float):
    def signal_handler(signum, frame):
        raise TimeoutException("Timed out!")
    signal.setitimer(signal.ITIMER_REAL, seconds)
    signal.signal(signal.SIGALRM, signal_handler)
    try:
        yield
    finally:
        signal.setitimer(signal.ITIMER_REAL, 0)


@contextlib.contextmanager
def swallow_io():
    stream = WriteOnlyStringIO()
    with contextlib.redirect_stdout(stream):
        with contextlib.redirect_stderr(stream):
            with redirect_stdin(stream):
                yield


@contextlib.contextmanager
def create_tempdir():
    with tempfile.TemporaryDirectory() as dirname:
        with chdir(dirname):
            yield dirname


class TimeoutException(Exception):
    pass


class WriteOnlyStringIO(io.StringIO):
    """ StringIO that throws an exception when it's read from """

    def read(self, *args, **kwargs):
        raise IOError

    def readline(self, *args, **kwargs):
        raise IOError

    def readlines(self, *args, **kwargs):
        raise IOError

    def readable(self, *args, **kwargs):
        """ Returns True if the IO object can be read. """
        return False


class redirect_stdin(contextlib._RedirectStream):  # type: ignore
    _stream = 'stdin'


@contextlib.contextmanager
def chdir(root):
    if root == ".":
        yield
        return
    cwd = os.getcwd()
    os.chdir(root)
    try:
        yield
    except BaseException as exc:
        raise exc
    finally:
        os.chdir(cwd)


def reliability_guard(maximum_memory_bytes: Optional[int] = None):
    """
    This disables various destructive functions and prevents the generated code
    from interfering with the test (e.g. fork bomb, killing other processes,
    removing filesystem files, etc.)

    WARNING
    This function is NOT a security sandbox. Untrusted code, including, model-
    generated code, should not be blindly executed outside of one. See the 
    Codex paper for more information about OpenAI's code sandbox, and proceed
    with caution.
    """

    if maximum_memory_bytes is not None:
        import resource
        resource.setrlimit(resource.RLIMIT_AS,
                           (maximum_memory_bytes, maximum_memory_bytes))
        resource.setrlimit(resource.RLIMIT_DATA,
                           (maximum_memory_bytes, maximum_memory_bytes))
        if not platform.uname().system == 'Darwin':
            resource.setrlimit(resource.RLIMIT_STACK,
                               (maximum_memory_bytes, maximum_memory_bytes))

    faulthandler.disable()

    import builtins
    builtins.exit = None
    builtins.quit = None

    import os
    os.environ['OMP_NUM_THREADS'] = '1'

    os.kill = None
    os.system = None
    os.putenv = None
    os.remove = None
    os.removedirs = None
    os.rmdir = None
    os.fchdir = None
    os.setuid = None
    os.fork = None
    os.forkpty = None
    os.killpg = None
    os.rename = None
    os.renames = None
    os.truncate = None
    os.replace = None
    os.unlink = None
    os.fchmod = None
    os.fchown = None
    os.chmod = None
    os.chown = None
    os.chroot = None
    os.fchdir = None
    os.lchflags = None
    os.lchmod = None
    os.lchown = None
    os.getcwd = None
    os.chdir = None

    import shutil
    shutil.rmtree = None
    shutil.move = None
    shutil.chown = None

    import subprocess
    subprocess.Popen = None  # type: ignore

    __builtins__['help'] = None

    import sys
    sys.modules['ipdb'] = None
    sys.modules['joblib'] = None
    sys.modules['resource'] = None
    sys.modules['psutil'] = None
    sys.modules['tkinter'] = None



"""Notion tool spec."""

from dotagent.tools.toolkits.base import BaseToolkit
from dotagent.tools.toolkits.notion_toolkit.notion.base import LoadData
from dotagent.tools.toolkits.notion_toolkit.notion.base import SearchData
from dotagent.tools.basetool import BaseTool
from typing import Optional, List

SEARCH_URL = "https://api.notion.com/v1/search"


class ZapierToolkit(BaseToolkit):
    api_key: str = None
    oauth_access_token: str = None

    def get_tools(self) -> List[BaseTool]:
        """Get the tools in the toolkit."""
        return [
            LoadData(api_key = self.api_key, oauth_access_token = self.oauth_access_token),
            SearchData(integration_token = self.api_key, oauth_access_token = self.oauth_access_token)
            ]


from typing import List

from dotagent.knowledgebase.document_loaders.basereader import BaseReader
from dotagent.schema import DocumentNode


class TrafilaturaWebReader(BaseReader):
    """Trafilatura web page reader.

    Reads pages from the web.
    Requires the `trafilatura` package.

    """

    def load_data(self, urls: List[str]) -> List[DocumentNode]:
        """Load data from the urls.

        Args:
            urls (List[str]): List of URLs to scrape.

        Returns:
            List[DocumentNode]: List of documents.

        """
        import trafilatura

        if not isinstance(urls, list):
            raise ValueError("urls must be a list of strings.")
        documents = []
        for url in urls:
            downloaded = trafilatura.fetch_url(url)
            response = trafilatura.extract(downloaded)
            metadata = {"url": url}
            documents.append(DocumentNode(text=response, extra_info=metadata))

        return documents



"""Slack toolkit."""

from dotagent.tools.toolkits.base import BaseToolkit
from dotagent.tools.basetool import BaseTool
from dotagent.tools.toolkits.slack_toolkit.slack_tool.utils import SlackReader
from dotagent.tools.toolkits.slack_toolkit.slack_tool.base import LoadData, FetchChannel, SendMessage
from typing import Optional, List
from ssl import SSLContext
import pydantic
import ssl
from pydantic import Field, validator
from datetime import datetime

class SlackToolkit(BaseToolkit):
    """Slack toolkit."""

    reader: Optional[SlackReader] = Field(None)
    slack_token: Optional[str] = Field(None)
    earliest_date: Optional[datetime] = Field(None)
    latest_date: Optional[datetime] = Field(None)

    class Config:
        arbitrary_types_allowed = True

    @validator('reader', pre=True, always=True)
    def set_reader(cls, v, values):
        # Create the SSLContext object here
        ssl_context = ssl.SSLContext()
        return SlackReader(
            slack_token=values.get('slack_token'),
            ssl=ssl_context,
            earliest_date=values.get('earliest_date'),
            latest_date=values.get('latest_date'),
        )

    def get_tools(self) -> List[BaseTool]:
        """Get the tools in the toolkit."""
        return [
            LoadData(
                    slack_token=self.slack_token,
                    ssl = self.ssl,
                    earliest_date = self.earliest_date,
                    latest_date = self.latest_date),
            FetchChannel(
                    slack_token=self.slack_token,
                    ssl = self.ssl,
                    earliest_date = self.earliest_date,
                    latest_date = self.latest_date),
            SendMessage(
                    slack_token=self.slack_token,
                    ssl = self.ssl,
                    earliest_date = self.earliest_date,
                    latest_date = self.latest_date),     
        ]


# class SlackToolkit(BaseToolkit):
#     """Slack toolkit."""

#     reader: Optional[SlackReader] = Field(None)

#     class Config:
#         arbitrary_types_allowed = True

#     def __init__(
#         self,
#         slack_token: Optional[str] = None,
#         ssl: Optional[SSLContext] = None,
#         earliest_date: Optional[datetime] = None,
#         latest_date: Optional[datetime] = None,
#     ) -> None:
#         """Initialize with parameters."""
#         self.reader: Optional[SlackReader] = None
#         self.reader = SlackReader(
#             slack_token=slack_token,
#             ssl=ssl,
#             earliest_date=earliest_date,
#             latest_date=latest_date,
#         )


# class SlackToolkit(BaseToolkit):
#     """Slack toolkit."""

#     reader: Optional[SlackReader] = Field(None)
#     slack_token: Optional[str] = Field(None)
#     ssl: Optional[SSLContext] = Field(None)
#     earliest_date: Optional[datetime] = Field(None)
#     latest_date: Optional[datetime] = Field(None)

#     class Config:
#         arbitrary_types_allowed = True

#     @pydantic.validator('reader', pre=True, always=True)
#     def set_reader(cls, v, values):
#         return SlackReader(
#             slack_token=values.get('slack_token'),
#             ssl=values.get('ssl'),
#             earliest_date=values.get('earliest_date'),
#             latest_date=values.get('latest_date'),
#          )


from .._utils import ContentCapture
from .._grammar import grammar

async def parse(string, name=None, hidden=False, _parser_context=None):
    ''' Parse a string as a Compiler program.

    This is useful for dynamically generating and then running Compiler programs (or parts of programs).

    Parameters
    ----------
    string : str
        The string to parse.
    name : str
        The name of the variable to set with the generated content.
    '''

    parser = _parser_context['parser']
    variable_stack = _parser_context['variable_stack']

    # capture the content of the block
    with ContentCapture(variable_stack, hidden) as new_content:

        # parse and visit the given string
        subtree = grammar.parse_string(string)
        new_content += await parser.visit(subtree, variable_stack)

        # save the content in a variable if needed
        if name is not None:
            variable_stack[name] = str(new_content)



"""Wrapper around Qdrant vector database."""
from __future__ import annotations

import uuid
import warnings
from itertools import islice
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

import numpy as np

from dotagent.schema import Document
from dotagent.vectorstores.embeddings.base import Embeddings
from dotagent.vectorstores import VectorStore

if TYPE_CHECKING:
    from qdrant_client.conversions import common_types
    from qdrant_client.http import models as rest

    DictFilter = Dict[str, Union[str, int, bool, dict, list]]
    MetadataFilter = Union[DictFilter, common_types.Filter]


class Qdrant(VectorStore):
    """Wrapper around Qdrant vector database.

    To use you should have the ``qdrant-client`` package installed.

    Example:
        .. code-block:: python

            from qdrant_client import QdrantClient
            from dotagent.vectorstores import Qdrant

            client = QdrantClient()
            collection_name = "MyCollection"
            qdrant = Qdrant(client, collection_name, embedding_function)
    """

    CONTENT_KEY = "page_content"
    METADATA_KEY = "metadata"

    def __init__(
        self,
        client: Any,
        collection_name: str,
        embedding_function: Optional[Callable] = None,
        content_payload_key: str = CONTENT_KEY,
        metadata_payload_key: str = METADATA_KEY,
    ):
        """Initialize with necessary components."""
        try:
            import qdrant_client
        except ImportError:
            raise ValueError(
                "Could not import qdrant-client python package. "
                "Please install it with `pip install qdrant-client`."
            )

        if not isinstance(client, qdrant_client.QdrantClient):
            raise ValueError(
                f"client should be an instance of qdrant_client.QdrantClient, "
                f"got {type(client)}"
            )

        self._embeddings_function = embedding_function
        self.client: qdrant_client.QdrantClient = client
        self.collection_name = collection_name
        self.content_payload_key = content_payload_key or self.CONTENT_KEY
        self.metadata_payload_key = metadata_payload_key or self.METADATA_KEY

        if embedding_function is not None:
            warnings.warn(
                "Using `embedding_function` is deprecated. "
                "Pass `Embeddings` instance to `embeddings` instead."
            )

        if not isinstance(embedding_function, Embeddings):
            warnings.warn(
                "`embeddings` should be an instance of `Embeddings`."
                "Using `embeddings` as `embedding_function` which is deprecated"
            )
            self._embeddings_function = embedding_function
            self.embeddings = None

    def add_texts(
        self,
        texts: Iterable[str],
        metadatas: Optional[List[dict]] = None,
        ids: Optional[Sequence[str]] = None,
        batch_size: int = 64,
        **kwargs: Any,
    ) -> List[str]:
        """Run more texts through the embeddings and add to the vectorstore.

        Args:
            texts: Iterable of strings to add to the vectorstore.
            metadatas: Optional list of metadatas associated with the texts.
            ids:
                Optional list of ids to associate with the texts. Ids have to be
                uuid-like strings.
            batch_size:
                How many vectors upload per-request.
                Default: 64

        Returns:
            List of ids from adding the texts into the vectorstore.
        """
        from qdrant_client.http import models as rest

        added_ids = []
        texts_iterator = iter(texts)
        metadatas_iterator = iter(metadatas or [])
        ids_iterator = iter(ids or [uuid.uuid4().hex for _ in iter(texts)])
        while batch_texts := list(islice(texts_iterator, batch_size)):
            # Take the corresponding metadata and id for each text in a batch
            batch_metadatas = list(islice(metadatas_iterator, batch_size)) or None
            batch_ids = list(islice(ids_iterator, batch_size))

            self.client.upsert(
                collection_name=self.collection_name,
                points=rest.Batch.construct(
                    ids=batch_ids,
                    vectors=self._embed_texts(batch_texts),
                    payloads=self._build_payloads(
                        batch_texts,
                        batch_metadatas,
                        self.content_payload_key,
                        self.metadata_payload_key,
                    ),
                ),
            )

            added_ids.extend(batch_ids)

        return added_ids
    


    def similarity_search(
        self,
        query: Optional[str],
        embedding: Optional[List[float]],
        k: int = 4,
        filter: Optional[MetadataFilter] = None,
        search_params: Optional[common_types.SearchParams] = None,
        offset: int = 0,
        score_threshold: Optional[float] = None,
        consistency: Optional[common_types.ReadConsistency] = None,
        **kwargs: Any,
    ) -> List[Document]:
        """Return docs most similar to query.

        Args:
            query: Text to look up documents similar to.
            k: Number of Documents to return. Defaults to 4.
            filter: Filter by metadata. Defaults to None.
            search_params: Additional search params
            offset:
                Offset of the first result to return.
                May be used to paginate results.
                Note: large offset values may cause performance issues.
            score_threshold:
                Define a minimal score threshold for the result.
                If defined, less similar results will not be returned.
                Score of the returned result might be higher or smaller than the
                threshold depending on the Distance function used.
                E.g. for cosine similarity only higher scores will be returned.
            consistency:
                Read consistency of the search. Defines how many replicas should be
                queried before returning the result.
                Values:
                - int - number of replicas to query, values should present in all
                        queried replicas
                - 'majority' - query all replicas, but return values present in the
                               majority of replicas
                - 'quorum' - query the majority of replicas, return values present in
                             all of them
                - 'all' - query all replicas, and return values present in all replicas

        Returns:
            List of Documents most similar to the query.
        """
        if (embedding is None and query is None) or (embedding is not None and query is not None):
            raise ValueError("You must provide either query embeddings or query texts, but not both")
        

        if filter is not None and isinstance(filter, dict):
            warnings.warn(
                "Using dict as a `filter` is deprecated. Please use qdrant-client "
                "filters directly: "
                "https://qdrant.tech/documentation/concepts/filtering/",
                DeprecationWarning,
            )
            qdrant_filter = self._qdrant_filter_from_dict(filter)
        else:
            qdrant_filter = filter
        results = self.client.search(
            collection_name=self.collection_name,
            query_vector=embedding,
            query_filter=qdrant_filter,
            search_params=search_params,
            limit=k,
            offset=offset,
            with_payload=True,
            with_vectors=False,  # dotagent does not expect vectors to be returned
            score_threshold=score_threshold,
            consistency=consistency,
            **kwargs,
        )
        return [
            (
                self._document_from_scored_point(
                    result, self.content_payload_key, self.metadata_payload_key
                ),
                result.score,
            )
            for result in results
        ]
    


    @classmethod
    def from_texts(
        cls: Type[Qdrant],
        texts: List[str],
        embedding_function: Embeddings,
        metadatas: Optional[List[dict]] = None,
        ids: Optional[Sequence[str]] = None,
        location: Optional[str] = None,
        batch_size: Optional[int] = 64,
        **kwargs: Any,
    ) -> Qdrant:
        """Construct Qdrant wrapper from a list of texts.

        Args:
            texts: A list of texts to be indexed in Qdrant.
            embedding: A subclass of `Embeddings`, responsible for text vectorization.
            metadatas:
                An optional list of metadata. If provided it has to be of the same
                length as a list of texts.
            ids:
                Optional list of ids to associate with the texts. Ids have to be
                uuid-like strings.
            location:
                If `:memory:` - use in-memory Qdrant instance.
                If `str` - use it as a `url` parameter.
                If `None` - fallback to relying on `host` and `port` parameters.
            **kwargs:
                Additional arguments passed directly into REST client initialization

        This is a user-friendly interface that:
        1. Creates embeddings, one for each text
        2. Initializes the Qdrant database as an in-memory docstore by default
           (and overridable to a remote docstore)
        3. Adds the text embeddings to the Qdrant database

        This is intended to be a quick way to get started.

        Example:
            .. code-block:: python

                from dotagent.vectorstores import Qdrant
                from dotagent.embeddings import OpenAIEmbeddings
                embeddings = OpenAIEmbeddings()
                qdrant = Qdrant.from_texts(texts, embeddings, "localhost")
        """
        try:
            import qdrant_client
        except ImportError:
            raise ValueError(
                "Could not import qdrant-client python package. "
                "Please install it with `pip install qdrant-client`."
            )

        from qdrant_client.http import models as rest

        # Just do a single quick embedding to get vector size
        partial_embeddings = embedding_function.embed_documents(texts[:1])
        vector_size = len(partial_embeddings[0])

        collection_name = collection_name or uuid.uuid4().hex
        distance_func = distance_func.upper()

        client = qdrant_client.QdrantClient(
            location=location,
            **kwargs,
        )

        client.recreate_collection(
            collection_name=collection_name,
            vectors_config=rest.VectorParams(
                size=vector_size,
                distance=rest.Distance[distance_func],
            ),
            timeout=timeout,  # type: ignore[arg-type]
        )

        texts_iterator = iter(texts)
        metadatas_iterator = iter(metadatas or [])
        ids_iterator = iter(ids or [uuid.uuid4().hex for _ in iter(texts)])
        while batch_texts := list(islice(texts_iterator, batch_size)):
            # Take the corresponding metadata and id for each text in a batch
            batch_metadatas = list(islice(metadatas_iterator, batch_size)) or None
            batch_ids = list(islice(ids_iterator, batch_size))

            # Generate the embeddings for all the texts in a batch
            batch_embeddings = embedding_function.embed_documents(batch_texts)

            client.upsert(
                collection_name=collection_name,
                points=rest.Batch.construct(
                    ids=batch_ids,
                    vectors=batch_embeddings,
                    payloads=cls._build_payloads(
                        batch_texts,
                        batch_metadatas,
                    ),
                ),
            )

        return cls(
            client=client,
            collection_name=collection_name,
            embedding_function=embedding_function,
        )

    @classmethod
    def _build_payloads(
        cls,
        texts: Iterable[str],
        metadatas: Optional[List[dict]],
        content_payload_key: str,
        metadata_payload_key: str,
    ) -> List[dict]:
        payloads = []
        for i, text in enumerate(texts):
            if text is None:
                raise ValueError(
                    "At least one of the texts is None. Please remove it before "
                    "calling .from_texts or .add_texts on Qdrant instance."
                )
            metadata = metadatas[i] if metadatas is not None else None
            payloads.append(
                {
                    content_payload_key: text,
                    metadata_payload_key: metadata,
                }
            )

        return payloads

    @classmethod
    def _document_from_scored_point(
        cls,
        scored_point: Any,
        content_payload_key: str,
        metadata_payload_key: str,
    ) -> Document:
        return Document(
            page_content=scored_point.payload.get(content_payload_key),
            metadata=scored_point.payload.get(metadata_payload_key) or {},
        )

    def _build_condition(self, key: str, value: Any) -> List[rest.FieldCondition]:
        from qdrant_client.http import models as rest

        out = []

        if isinstance(value, dict):
            for _key, value in value.items():
                out.extend(self._build_condition(f"{key}.{_key}", value))
        elif isinstance(value, list):
            for _value in value:
                if isinstance(_value, dict):
                    out.extend(self._build_condition(f"{key}[]", _value))
                else:
                    out.extend(self._build_condition(f"{key}", _value))
        else:
            out.append(
                rest.FieldCondition(
                    key=f"{self.metadata_payload_key}.{key}",
                    match=rest.MatchValue(value=value),
                )
            )

        return out

    def _qdrant_filter_from_dict(
        self, filter: Optional[DictFilter]
    ) -> Optional[rest.Filter]:
        from qdrant_client.http import models as rest

        if not filter:
            return None

        return rest.Filter(
            must=[
                condition
                for key, value in filter.items()
                for condition in self._build_condition(key, value)
            ]
        )



from os import PathLike
from typing import Any, Dict, List, Optional
from uuid import UUID
from aiohttp import ClientSession
from dotagent.agent.agentbox.basebox import BaseBox
from dotagent.config import settings
from dotagent.schema import AgentBoxFile, AgentBoxOutput, AgentBoxStatus
from dotagent.agent.agentbox._utils import abase_request, base_request


class AgentBox(BaseBox):
    """
    Agentboxed Python Interpreter
    """

    def __new__(cls, *args, **kwargs):
        if (
            kwargs.pop("local", False)
            or settings.AGENTBOX_API_KEY is None
            or settings.AGENTBOX_API_KEY == "local"
        ):
            from .tinybox import TinyBox

            return TinyBox(*args, **kwargs)

        return super().__new__(cls, *args, **kwargs)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.session_id: Optional[UUID] = kwargs.get("id", None)
        self.aiohttp_session: Optional[ClientSession] = None

    def agentbox_request(self, method, endpoint, *args, **kwargs) -> Dict[str, Any]:
        """Basic request to the AgentBox API"""
        self._update()
        return base_request(
            method, f"/agentbox/{self.session_id}" + endpoint, *args, **kwargs
        )

    async def aagentbox_request(
        self, method, endpoint, *args, **kwargs
    ) -> Dict[str, Any]:
        """Basic async request to the AgentBox API"""
        self._update()
        if self.aiohttp_session is None:
            raise RuntimeError("AgentBox session not started")
        return await abase_request(
            self.aiohttp_session,
            method,
            f"/agentbox/{self.session_id}" + endpoint,
            *args,
            **kwargs,
        )

    def start(self) -> AgentBoxStatus:
        self.session_id = base_request(
            method="GET",
            endpoint="/agentbox/start",
        )["id"]
        return AgentBoxStatus(status="started")

    async def astart(self) -> AgentBoxStatus:
        self.aiohttp_session = ClientSession()
        self.session_id = (
            await abase_request(
                self.aiohttp_session,
                method="GET",
                endpoint="/agentbox/start",
            )
        )["id"]
        return AgentBoxStatus(status="started")

    def status(self):
        return AgentBoxStatus(
            **self.agentbox_request(
                method="GET",
                endpoint="/",
            )
        )

    async def astatus(self):
        return AgentBoxStatus(
            **await self.aagentbox_request(
                method="GET",
                endpoint="/",
            )
        )

    def run(
        self, code: Optional[str] = None, file_path: Optional[PathLike] = None
    ) -> AgentBoxOutput:
        if not code and not file_path:  # R0801
            raise ValueError("Code or file_path must be specified!")

        if code and file_path:
            raise ValueError("Can only specify code or the file to read_from!")

        if file_path:
            with open(file_path, "r", encoding="utf-8") as f:
                code = f.read()

        return AgentBoxOutput(
            **self.agentbox_request(
                method="POST",
                endpoint="/run",
                body={"code": code},
            )
        )

    async def arun(
        self, code: str, file_path: Optional[PathLike] = None
    ) -> AgentBoxOutput:
        if file_path:  # TODO: Implement this
            raise NotImplementedError(
                "Reading from FilePath is not supported in async mode yet!"
            )

        return AgentBoxOutput(
            **await self.aagentbox_request(
                method="POST",
                endpoint="/run",
                body={"code": code},
            )
        )

    def upload(self, file_name: str, content: bytes) -> AgentBoxStatus:
        return AgentBoxStatus(
            **self.agentbox_request(
                method="POST",
                endpoint="/upload",
                files={"file": (file_name, content)},
            )
        )

    async def aupload(self, file_name: str, content: bytes) -> AgentBoxStatus:
        return AgentBoxStatus(
            **await self.aagentbox_request(
                method="POST",
                endpoint="/upload",
                files={"file": (file_name, content)},
            )
        )

    def download(self, file_name: str) -> AgentBoxFile:
        return AgentBoxFile(
            **self.agentbox_request(
                method="GET",
                endpoint="/download",
                body={"file_name": file_name},
            )
        )

    async def adownload(self, file_name: str) -> AgentBoxFile:
        return AgentBoxFile(
            **await self.aagentbox_request(
                method="GET",
                endpoint="/download",
                body={"file_name": file_name},
            )
        )

    def install(self, package_name: str) -> AgentBoxStatus:
        return AgentBoxStatus(
            **self.agentbox_request(
                method="POST",
                endpoint="/install",
                body={
                    "package_name": package_name,
                },
            )
        )

    async def ainstall(self, package_name: str) -> AgentBoxStatus:
        return AgentBoxStatus(
            **await self.aagentbox_request(
                method="POST",
                endpoint="/install",
                body={
                    "package_name": package_name,
                },
            )
        )

    def list_files(self) -> List[AgentBoxFile]:
        return [
            AgentBoxFile(name=file_name, content=None)
            for file_name in (
                self.agentbox_request(
                    method="GET",
                    endpoint="/files",
                )
            )["files"]
        ]

    async def alist_files(self) -> List[AgentBoxFile]:
        return [
            AgentBoxFile(name=file_name, content=None)
            for file_name in (
                await self.aagentbox_request(
                    method="GET",
                    endpoint="/files",
                )
            )["files"]
        ]

    def restart(self) -> AgentBoxStatus:
        return AgentBoxStatus(
            **self.agentbox_request(
                method="POST",
                endpoint="/restart",
            )
        )

    async def arestart(self) -> AgentBoxStatus:
        return AgentBoxStatus(
            **await self.aagentbox_request(
                method="POST",
                endpoint="/restart",
            )
        )

    def stop(self) -> AgentBoxStatus:
        return AgentBoxStatus(
            **self.agentbox_request(
                method="POST",
                endpoint="/stop",
            )
        )

    async def astop(self) -> AgentBoxStatus:
        status = AgentBoxStatus(
            **await self.aagentbox_request(
                method="POST",
                endpoint="/stop",
            )
        )
        if self.aiohttp_session:
            await self.aiohttp_session.close()
            self.aiohttp_session = None
        return status



import asyncio
from typing import Any, Dict

import pytest

from dotagent.tools.toolkits.requests_toolkit.requests.utils import TextRequestsWrapper
from dotagent.tools.toolkits.requests_toolkit.requests.base import (
    RequestsDeleteTool,
    RequestsGetTool,
    RequestsPatchTool,
    RequestsPostTool,
    RequestsPutTool,
    _parse_input,
)


class _MockTextRequestsWrapper(TextRequestsWrapper):
    @staticmethod
    def get(url: str, **kwargs: Any) -> str:
        return "get_response"

    @staticmethod
    def post(url: str, data: Dict[str, Any], **kwargs: Any) -> str:
        return f"post {str(data)}"


    @staticmethod
    def patch(url: str, data: Dict[str, Any], **kwargs: Any) -> str:
        return f"patch {str(data)}"

    @staticmethod
    def put(url: str, data: Dict[str, Any], **kwargs: Any) -> str:
        return f"put {str(data)}"

    @staticmethod
    def delete(url: str, **kwargs: Any) -> str:
        return "delete_response"

@pytest.fixture
def mock_requests_wrapper() -> TextRequestsWrapper:
    return _MockTextRequestsWrapper()


def test_parse_input() -> None:
    input_text = '{"url": "https://example.com", "data": {"key": "value"}}'
    expected_output = {"url": "https://example.com", "data": {"key": "value"}}
    assert _parse_input(input_text) == expected_output


def test_requests_get_tool(mock_requests_wrapper: TextRequestsWrapper) -> None:
    tool = RequestsGetTool(requests_wrapper=mock_requests_wrapper)
    assert tool.run("https://example.com") == "get_response"


def test_requests_post_tool(mock_requests_wrapper: TextRequestsWrapper) -> None:
    tool = RequestsPostTool(requests_wrapper=mock_requests_wrapper)
    input_text = '{"url": "https://example.com", "data": {"key": "value"}}'
    assert tool.run(input_text) == "post {'key': 'value'}"


def test_requests_patch_tool(mock_requests_wrapper: TextRequestsWrapper) -> None:
    tool = RequestsPatchTool(requests_wrapper=mock_requests_wrapper)
    input_text = '{"url": "https://example.com", "data": {"key": "value"}}'
    assert tool.run(input_text) == "patch {'key': 'value'}"


def test_requests_put_tool(mock_requests_wrapper: TextRequestsWrapper) -> None:
    tool = RequestsPutTool(requests_wrapper=mock_requests_wrapper)
    input_text = '{"url": "https://example.com", "data": {"key": "value"}}'
    assert tool.run(input_text) == "put {'key': 'value'}"


def test_requests_delete_tool(mock_requests_wrapper: TextRequestsWrapper) -> None:
    tool = RequestsDeleteTool(requests_wrapper=mock_requests_wrapper)
    assert tool.run("https://example.com") == "delete_response"


from typing import Dict, Optional, Any
import json

from dotagent.vectorstores.chroma import Chroma
from dotagent.vectorstores.embeddings.base import Embeddings

from . import BaseCache

class ChromaSemanticCache(BaseCache):
    """ChromaSemanticCache is a semantic cache that uses Chroma for caching."""

    def __init__(
        self,
        collection_name: str = "cache_collection",
        threshold: float = 0.2,
        embedding_function: Optional[Embeddings] = None
    ):
        self._cache_collection = Chroma(
            collection_name=collection_name, embedding_function=embedding_function
        )
        self._threshold = threshold

    def __getitem__(self, key: str) -> Any:
        """Get an item from the cache or throw key error."""
        response = self._cache_collection.similarity_search(query=key, top_k=1)
        if response and response[0][1] < self._threshold:
            # Deserialize JSON string to Python object
            return json.loads(response[0][0].metadata)
        else:
            raise KeyError(f"No data found for key: {key}")

    def __setitem__(self, key: str, value: Any) -> None:
        """Set an item in the cache."""
        # Serialize Python object to JSON string
        metadata = json.dumps({"prompt": key, "answer": value})
        self._cache_collection.add_texts(texts=[key], metadatas=[metadata])

    def __contains__(self, key: str) -> bool:
        """Check if a key is in the cache."""
        response = self._cache_collection.similarity_search(query=key, top_k=1)
        return response and response[0][1] < self._threshold

    def create_key(self, llm: str, **kwargs: Dict[str, Any]) -> str:
        """Create a lookup key for a call to the given llm with the given kwargs."""
        if "cache_key" in kwargs:
            return str(kwargs["cache_key"])
        elif "prompt" in kwargs:
            return str(kwargs["prompt"])
        else:
            raise ValueError("Expected 'cache_key' or 'prompt' in kwargs")

    def clear(self) -> None:
        """Clear the cache."""
        self._cache_collection.delete_collection()



"""Notion reader."""
import os
from typing import Any, Dict, List, Optional

import requests  # type: ignore
from dotagent.knowledgebase.document_loaders.basereader import BaseReader
from dotagent.schema import DocumentNode

INTEGRATION_TOKEN_NAME = "NOTION_INTEGRATION_TOKEN"
BLOCK_CHILD_URL_TMPL = "https://api.notion.com/v1/blocks/{block_id}/children"
DATABASE_URL_TMPL = "https://api.notion.com/v1/databases/{database_id}/query"
SEARCH_URL = "https://api.notion.com/v1/search"


# TODO: Notion DB reader coming soon!
class NotionPageReader(BaseReader):
    """Notion Page reader.

    Reads a set of Notion pages.

    Args:
        integration_token (str): Notion integration token.

    """

    def __init__(self, integration_token: Optional[str] = None) -> None:
        """Initialize with parameters."""
        if integration_token is None:
            integration_token = os.getenv(INTEGRATION_TOKEN_NAME)
            if integration_token is None:
                raise ValueError(
                    "Must specify `integration_token` or set environment "
                    "variable `NOTION_INTEGRATION_TOKEN`."
                )
        self.token = integration_token
        self.headers = {
            "Authorization": "Bearer " + self.token,
            "Content-Type": "application/json",
            "Notion-Version": "2022-06-28",
        }

    def _read_block(self, block_id: str, num_tabs: int = 0) -> str:
        """Read a block."""
        done = False
        result_lines_arr = []
        cur_block_id = block_id
        while not done:
            block_url = BLOCK_CHILD_URL_TMPL.format(block_id=cur_block_id)
            query_dict: Dict[str, Any] = {}

            res = requests.request(
                "GET", block_url, headers=self.headers, json=query_dict
            )
            data = res.json()

            for result in data["results"]:
                result_type = result["type"]
                result_obj = result[result_type]

                cur_result_text_arr = []
                if "rich_text" in result_obj:
                    for rich_text in result_obj["rich_text"]:
                        # skip if doesn't have text object
                        if "text" in rich_text:
                            text = rich_text["text"]["content"]
                            prefix = "\t" * num_tabs
                            cur_result_text_arr.append(prefix + text)

                result_block_id = result["id"]
                has_children = result["has_children"]
                if has_children:
                    children_text = self._read_block(
                        result_block_id, num_tabs=num_tabs + 1
                    )
                    cur_result_text_arr.append(children_text)

                cur_result_text = "\n".join(cur_result_text_arr)
                result_lines_arr.append(cur_result_text)

            if data["next_cursor"] is None:
                done = True
                break
            else:
                cur_block_id = data["next_cursor"]

        result_lines = "\n".join(result_lines_arr)
        return result_lines

    def read_page(self, page_id: str) -> str:
        """Read a page."""
        return self._read_block(page_id)

    def query_database(
        self, database_id: str, query_dict: Dict[str, Any] = {"page_size": 100}
    ) -> List[str]:
        """Get all the pages from a Notion database."""
        pages = []

        res = requests.post(
            DATABASE_URL_TMPL.format(database_id=database_id),
            headers=self.headers,
            json=query_dict,
        )
        res.raise_for_status()
        data = res.json()

        pages.extend(data.get("results"))

        while data.get("has_more"):
            query_dict["start_cursor"] = data.get("next_cursor")
            res = requests.post(
                DATABASE_URL_TMPL.format(database_id=database_id),
                headers=self.headers,
                json=query_dict,
            )
            res.raise_for_status()
            data = res.json()
            pages.extend(data.get("results"))

        page_ids = [page["id"] for page in pages]
        return page_ids

    def search(self, query: str) -> List[str]:
        """Search Notion page given a text query."""
        done = False
        next_cursor: Optional[str] = None
        page_ids = []
        while not done:
            query_dict = {
                "query": query,
            }
            if next_cursor is not None:
                query_dict["start_cursor"] = next_cursor
            res = requests.post(SEARCH_URL, headers=self.headers, json=query_dict)
            data = res.json()
            for result in data["results"]:
                page_id = result["id"]
                page_ids.append(page_id)

            if data["next_cursor"] is None:
                done = True
                break
            else:
                next_cursor = data["next_cursor"]
        return page_ids

    def load_data(
        self, page_ids: List[str] = [], database_id: Optional[str] = None
    ) -> List[DocumentNode]:
        """Load data from the input directory.

        Args:
            page_ids (List[str]): List of page ids to load.
            database_id (str): Database_id from which to load page ids.

        Returns:
            List[DocumentNode]: List of documents.

        """
        if not page_ids and not database_id:
            raise ValueError("Must specify either `page_ids` or `database_id`.")
        docs = []
        if database_id is not None:
            # get all the pages in the database
            page_ids = self.query_database(database_id)
            for page_id in page_ids:
                page_text = self.read_page(page_id)
                docs.append(DocumentNode(text=page_text, extra_info={"page_id": page_id, "database_id": database_id}))
        else:
            for page_id in page_ids:
                page_text = self.read_page(page_id)
                docs.append(DocumentNode(text=page_text, extra_info={"page_id": page_id, "database_id": database_id}))

        return docs


if __name__ == "__main__":
    reader = NotionPageReader()
    print(reader.search("What I"))



from typing import Any

from . import BaseCache

class LocalMemoryCache(BaseCache):
    def __init__(self) -> None:
        """Initialize an empty memory cache."""
        self._memory_cache = {}

    def __getitem__(self, key: str) -> Any:
        """Get the value for a key. Raise KeyError if the key is not found."""
        try:
            return self._memory_cache[key]
        except KeyError:
            raise KeyError(f"Key {key} not found in cache.")

    def __setitem__(self, key: str, value: Any) -> None:
        """Set a value for a key in the memory cache."""
        self._memory_cache[key] = value

    def __contains__(self, key: str) -> bool:
        """Check if a key is in the memory cache."""
        return key in self._memory_cache

    def clear(self) -> None:
        """Clear the memory cache."""
        self._memory_cache = {}



'''
Env values will be passed from from os.environ
Automatically loads environment variables from .env file
'''

from typing import Optional
import os
import yaml
from pathlib import Path
from dotenv import load_dotenv
from pydantic import BaseSettings
load_dotenv("./.env")


CONFIG_FILE = "config.yaml"
ROOT_DIR = os.path.dirname(Path(__file__).parent.parent)
config_path = ROOT_DIR + "/" + CONFIG_FILE

if os.path.exists(config_path):
    with open(config_path, "r") as file:
        config_data = yaml.safe_load(file)

else:
    config_data = {'OPENAI_API_KEY': ""}
    config_data['OPENAI_API_KEY'] = input("Enter OPENAI_API_KEY:")
    config_data["OPENAI_ORG_ID"] = input("Enter OPENAI_ORG_ID:")
    config_data["SERP_API_KEY"] = input("Enter SERP_API_KEY:")
    config_data["GOOGLE_SEARCH_API_KEY"] = input("Enter GOOGLE_SEARCH_API_KEY:")


class Config:
    OPENAI_API_KEY = config_data["OPENAI_API_KEY"]
    OPENAI_ORG_ID = config_data["OPENAI_ORG_ID"]
    SERP_API_KEY = config_data["SERP_API_KEY"]
    GOOGLE_SEARCH_API_KEY = config_data["GOOGLE_SEARCH_API_KEY"]


class AgentBoxSettings(BaseSettings):
    """
    AgentBox API Config
    """

    VERBOSE: bool = False
    SHOW_INFO: bool = True

    AGENTBOX_API_KEY: Optional[str] = None
    AGENTBOX_BASE_URL: str = "https://agentboxapi.com/api/v1"
    AGENTBOX_TIMEOUT: int = 20


settings = AgentBoxSettings()


"""CouchDB client."""

from typing import Dict, List, Optional
from dotagent.knowledgebase.document_loaders.basereader import BaseReader
from dotagent.schema import DocumentNode
import logging
import json


class SimpleCouchDBReader(BaseReader):
    """Simple CouchDB reader.

    Concatenates each CouchDB doc into DocumentNode used by LlamaIndex.

    Args:
        couchdb_url (str): CouchDB Full URL.
        max_docs (int): Maximum number of documents to load.

    """

    def __init__(
        self,
        user: str,
        pwd: str,
        host: str,
        port: int,
        couchdb_url: Optional[Dict] = None,
        max_docs: int = 1000,
    ) -> None:
        """Initialize with parameters."""

        self.user = user

        import couchdb3

        if couchdb_url is not None:
            self.client: CouchDBClient = couchdb3.Server(couchdb_url)
        else:
            self.client: CouchDBClient = couchdb3.Server(
                f"http://{user}:{pwd}@{host}:{port}"
            )
        self.max_docs = max_docs

    def load_data(self, db_name: str, query: Optional[str] = None) -> List[DocumentNode]:
        """Load data from the input directory.

        Args:
            db_name (str): name of the database.
            query (Optional[str]): query to filter documents.
                Defaults to None

        Returns:
            List[DocumentNode]: A list of documents.

        """

        metadata = {
            "user": self.user,
            "db_name": db_name,
            "query": query
        }

        documents = []
        db = self.client.get(db_name)
        if query is None:
            # if no query is specified, return all docs in database
            logging.debug("showing all docs")
            results = db.view("_all_docs", include_docs=True)
        else:
            logging.debug("executing query")
            results = db.find(query)

        if type(results) is not dict:
            logging.debug(results.rows)
        else:
            logging.debug(results)

        # check if more than one result
        if type(results) is not dict and results.rows is not None:
            for row in results.rows:
                # check that the id field exists
                if "id" not in row:
                    raise ValueError("`id` field not found in CouchDB DocumentNode.")
                documents.append(DocumentNode(text=json.dumps(row.doc), extra_info=metadata))
        else:
            # only one result
            if results.get("docs") is not None:
                for item in results.get("docs"):
                    # check that the _id field exists
                    if "_id" not in item:
                        raise ValueError("`_id` field not found in CouchDB DocumentNode.")
                    documents.append(DocumentNode(text=json.dumps(item), extra_info=metadata))

        return documents



import builtins

def callable(value, _parser_context=None):
    ''' Check if the given value is callable.
    '''
    variable_stack = _parser_context["variable_stack"]

    function_call = variable_stack["llm.extract_function_call"](value)
    if function_call is not None:
        return True

    return builtins.callable(value)


import uuid

from dotagent.schema import Document
from dotagent.vectorstores.base import VectorStore
from typing import Any, Optional, Iterable, List, Dict
from dotagent.vectorstores.embeddings.base import Embeddings
import chromadb
import chromadb.config

class Chroma(VectorStore):
    """Wrapper around ChromaDB embeddings platform.

    To use, you should have the ``chromadb`` python package installed.

    Example:
        .. code-block:: python

                from dotagent.vectorstores import Chroma
                from dotagent.embeddings.openai import OpenAIEmbeddings

                embeddings = OpenAIEmbeddings()
                vectorstore = Chroma("dotagent_collection", embeddings)
    """
    
    def __init__(
        self,
        collection_name: str = "dotagent",
        embedding_function: Optional[Embeddings] = None,
        persist_directory: Optional[str] = None,
        client_settings: Optional[chromadb.config.Settings] = None,
        collection_metadata: Optional[Dict] = None,
        client: Optional[chromadb.Client] = None,
    ) -> None:
        """Initialize with Chroma client."""

        try:
            import chromadb
            import chromadb.config
        except ImportError:
            raise ValueError(
                "Could not import chromadb python package. "
                "Please install it with `pip install chromadb`."
            )
        
        if client is not None:
            self._client = client
        else:
            if client_settings:
                self._client_settings = client_settings
            else:
                self._client_settings = chromadb.config.Settings()
                if persist_directory is not None:
                    self._client_settings = chromadb.config.Settings(
                        chroma_db_impl="duckdb+parquet",
                        persist_directory=persist_directory,
                    )
            self._client = chromadb.Client(self._client_settings)

        self._embedding_function = embedding_function
        self._persist_directory = persist_directory
        self._collection = self._client.get_or_create_collection(
            name=collection_name,
            embedding_function=self._embedding_function.embed_documents
            if self._embedding_function is not None
            else None,
            metadata=collection_metadata,
        )

    def add_texts(
        self,
        texts: Iterable[str],
        metadatas: Optional[List[dict]] = None,
        ids: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> List[str]:
        
        ids = ids or [str(uuid.uuid1()) for _ in texts]
        embeddings = None
        if self._embedding_function is not None:
            embeddings = self._embedding_function.embed_documents(list(texts))
        self._collection.add(
            metadatas=metadatas, embeddings=embeddings, documents=texts, ids=ids
        )
        return ids

    def similarity_search(
            self,
            query: Optional[str] = None, 
            embedding: Optional[List[float]] = None, 
            top_k: int = 5, 
            **kwargs: Any
            ) -> List[Document]:
        """Return docs most similar to query and respective distance score"""

        if (embedding is None and query is None) or (embedding is not None and query is not None):
            raise ValueError("You must provide either query embeddings or query texts, but not both")

        search_results = []
        if self._embedding_function is None:
            if query is not None:
                results = self._collection.query(
                    query_texts=[query],
                    n_results=top_k,
                    **kwargs,
                )
            else:
                results = self._collection.query(
                    query_embeddings=[embedding],
                    n_results=top_k,
                    **kwargs,
                )
        else:
            if query is not None:
                query_embedding = self._embedding_function.embed_query(text=query)
                results = self._collection.query(
                    query_embeddings=[query_embedding],
                    n_results=top_k,
                    **kwargs,
                )
            else:
                results = self._collection.query(
                    query_embeddings=[embedding],
                    n_results=top_k,
                    **kwargs,
                )

        for result in zip(
            results["documents"][0],
            results["metadatas"][0],
            results["distances"][0],
        ):
            document = Document(page_content=result[0], metadata=result[1] or {})
            search_results.append((document, result[2]))

        return search_results
    
    def delete_collection(self) -> None:
        """Delete the collection."""
        self._client.delete_collection(self._collection.name)

    def get(self, include: Optional[List[str]] = None) -> Dict[str, Any]:
        """Gets the collection.

        Args:
            include (Optional[List[str]]): List of fields to include from db.
                Defaults to None.
        """
        if include is not None:
            return self._collection.get(include=include)
        else:
            return self._collection.get()

    def persist(self) -> None:
        """Persist the collection.

        This can be used to explicitly persist the data to disk.
        It will also be called automatically when the object is destroyed.
        """
        if self._persist_directory is None:
            raise ValueError(
                "You must specify a persist_directory on"
                "creation to persist the collection."
            )
        self._client.persist()
    
    def update_document(self, document_id: str, document: Document) -> None:
        # Update a document in the collection.

        text = document.page_content
        metadata = document.metadata
        if self._embedding_function is None:
            raise ValueError(
                "For update, you must specify an embedding function on creation."
            )
        embeddings = self._embedding_function.get_document_embedding([text])

        self._collection.update(
            ids=[document_id],
            embeddings=embeddings,
            documents=[text],
            metadatas=[metadata],
        )

    @classmethod
    def from_texts(
        cls,
        texts: List[str],
        embedding_function: Optional[Embeddings] = None,
        metadatas: Optional[List[dict]] = None,
        ids: Optional[List[str]] = None,
        collection_name: str = "dotagent",
        persist_directory: Optional[str] = None,
        client_settings: Optional[chromadb.config.Settings] = None,
        client: Optional[chromadb.Client] = None,
        **kwargs: Any,
    ):
        """Create a Chroma vectorstore from a raw documents.

        If a persist_directory is specified, the collection will be persisted there.
        Otherwise, the data will be ephemeral in-memory.

        Args:
            texts (List[str]): List of texts to add to the collection.
            collection_name (str): Name of the collection to create.
            persist_directory (Optional[str]): Directory to persist the collection.
            embedding (Optional[Embeddings]): Embedding function. Defaults to None.
            metadatas (Optional[List[dict]]): List of metadatas. Defaults to None.
            ids (Optional[List[str]]): List of document IDs. Defaults to None.
            client_settings (Optional[chromadb.config.Settings]): Chroma client settings

        Returns:
            Chroma: Chroma vectorstore.
        """
        chroma_collection = cls(
            collection_name=collection_name,
            embedding_function=embedding_function,
            persist_directory=persist_directory,
            client_settings=client_settings,
            client=client,
        )
        chroma_collection.add_texts(texts=texts, metadatas=metadatas, ids=ids)
        return chroma_collection


    @classmethod
    def from_documents(
        cls,
        documents: List[Document],
        embedding_function: Optional[Embeddings] = None,
        ids: Optional[List[str]] = None,
        collection_name: str = "dotagent",
        persist_directory: Optional[str] = None,
        client_settings: Optional[chromadb.config.Settings] = None,
        client: Optional[chromadb.Client] = None,  # Add this line
        **kwargs: Any,
    ):
        """Create a Chroma vectorstore from a list of documents.

        If a persist_directory is specified, the collection will be persisted there.
        Otherwise, the data will be ephemeral in-memory.

        Args:
            collection_name (str): Name of the collection to create.
            persist_directory (Optional[str]): Directory to persist the collection.
            ids (Optional[List[str]]): List of document IDs. Defaults to None.
            documents (List[Document]): List of documents to add to the vectorstore.
            embedding (Optional[Embeddings]): Embedding function. Defaults to None.
            client_settings (Optional[chromadb.config.Settings]): Chroma client settings
        Returns:
            Chroma: Chroma vectorstore.
        """
        texts = [doc.page_content for doc in documents]
        metadatas = [doc.metadata for doc in documents]
        return cls.from_texts(
            texts=texts,
            embedding=embedding_function,
            metadatas=metadatas,
            ids=ids,
            collection_name=collection_name,
            persist_directory=persist_directory,
            client_settings=client_settings,
            client=client,
            **kwargs
        )


from transformers import AutoModelForCausalLM, AutoTokenizer, Trainer, TrainingArguments
from datasets import load_dataset
from logging import Logger
import openai
from . import LLMFinetune

class TransformersFinetune(LLMFinetune):
    def __init__(self, logger: Logger, base_model: str):
        super().__init__(logger, openai_key=None)
        self.model = AutoModelForCausalLM.from_pretrained(base_model)
        self.tokenizer = AutoTokenizer.from_pretrained(base_model)

    def transform_data(self, train_csv_file: str, val_csv_file: str, train_output_file: str, val_output_file: str) -> str:
        # Implement logic to transform CSV files to desired JSON or other formats
        # You can load, process, and save the CSV data here
        # Return the path or message confirming the transformation
        pass

    def finetune(self, data_path, output_dir, num_epochs=1, batch_size=32, learning_rate=5e-5, val_set_size=0.1, max_length=512):
        # Load dataset
        data = load_dataset("json", data_files={"train": data_path})

        # Split data into training and validation sets
        train_val = data["train"].train_test_split(test_size=val_set_size, shuffle=True, seed=42)
        train_data = train_val["train"]
        valid_data = train_val["test"]

        # Tokenization function
        def tokenize_function(examples):
            return self.tokenizer(examples["text"], truncation=True, max_length=max_length, padding="max_length")

        # Tokenize dataset
        train_data = train_data.map(tokenize_function, batched=True)
        valid_data = valid_data.map(tokenize_function, batched=True)

        # Training arguments
        training_args = TrainingArguments(
            per_device_train_batch_size=batch_size,
            per_device_eval_batch_size=batch_size,
            num_train_epochs=num_epochs,
            learning_rate=learning_rate,
            output_dir=output_dir,
            evaluation_strategy="steps" if val_set_size > 0 else "no",
            logging_dir='./logs',
        )

        # Trainer
        trainer = Trainer(
            model=self.model,
            args=training_args,
            train_dataset=train_data,
            eval_dataset=valid_data,
        )

        # Training
        trainer.train()

        # Save model
        self.model.save_pretrained(output_dir)



import json
import numpy as np

class CustomJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.int64):
            return int(obj)
        elif isinstance(obj, tuple):
            return list(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return super(CustomJSONEncoder, self).default(obj)


"""Firebase Realtime Database Loader."""

from typing import Dict, List, Optional

from dotagent.knowledgebase.document_loaders.basereader import BaseReader
from dotagent.schema import DocumentNode


class FirebaseRealtimeDatabaseReader(BaseReader):
    """Firebase Realtime Database reader.

    Retrieves data from Firebase Realtime Database and converts it into the DocumentNode used by LlamaIndex.

    Args:
        database_url (str): Firebase Realtime Database URL.
        service_account_key_path (Optional[str]): Path to the service account key file.

    """

    def __init__(
        self,
        database_url: str,
        service_account_key_path: Optional[str] = None,
    ) -> None:
        """Initialize with parameters."""

        self.database_url = database_url

        try:
            import firebase_admin
            from firebase_admin import credentials
        except ImportError:
            raise ImportError("`firebase_admin` package not found, please run `pip install firebase-admin`")
        
        if not firebase_admin._apps:
            if service_account_key_path:
                cred = credentials.Certificate(service_account_key_path)
                firebase_admin.initialize_app(
                    cred, options={"databaseURL": database_url}
                )
            else:
                firebase_admin.initialize_app(
                    options={"databaseURL": database_url}
                )
        

    def load_data(self, path: str, field: Optional[str] = None) -> List[DocumentNode]:
        """Load data from Firebase Realtime Database and convert it into documents.

        Args:
            path (str): Path to the data in the Firebase Realtime Database.
            field (str, Optional): Key to pick data from

        Returns:
            List[DocumentNode]: A list of documents.

        """

        try:
            from firebase_admin import db
        except ImportError:
            raise ImportError("`firebase_admin` package not found, please run `pip install firebase-admin`")
        
        ref = db.reference(path)
        data = ref.get()

        documents = []

        if isinstance(data, Dict):
            for key in data:
                entry = data[key]
                metadata = {
                    "document_id": key,
                    "databaseURL": self.database_url,
                    "path": path,
                    "field": field 
                }
                if type(entry) is Dict and field in entry:
                  text = entry[field]
                else:
                  text = str(entry)
                
                DocumentNode = DocumentNode(text=text, extra_info=metadata)
                documents.append(DocumentNode)
        elif isinstance(data, str):
            documents.append(DocumentNode(text=data))

        return documents


"""Generic utility functions."""
import os
from typing import Any, Dict, Optional,List


def get_from_dict_or_env(
    data: Dict[str, Any], key: str, env_key: str, default: Optional[str] = None
) -> str:
    """Get a value from a dictionary or an environment variable."""
    if key in data and data[key]:
        return data[key]
    else:
        return get_from_env(key, env_key, default=default)


def get_from_env(key: str, env_key: str, default: Optional[str] = None) -> str:
    """Get a value from a dictionary or an environment variable."""
    if env_key in os.environ and os.environ[env_key]:
        return os.environ[env_key]
    elif default is not None:
        return default
    else:
        raise ValueError(
            f"Did not find {key}, please add an environment variable"
            f" `{env_key}` which contains it, or pass"
            f"  `{key}` as a named parameter."
        )


def stringify_value(val: Any) -> str:
    if isinstance(val, str):
        return val
    elif isinstance(val, dict):
        return "\n" + stringify_dict(val)
    elif isinstance(val, list):
        return "\n".join(stringify_value(v) for v in val)
    else:
        return str(val)


def stringify_dict(data: dict) -> str:
    text = ""
    for key, value in data.items():
        text += key + ": " + stringify_value(value) + "\n"
    return text

def comma_list(items: List[Any]) -> str:
    return ", ".join(str(item) for item in items)



import logging
import copy
import pdb
import math
import os
import json
import yaml
import time
import re
from typing import List, Dict

from dotagent.fact_check.utils.base.pipeline import pipeline
from dotagent.fact_check.code.helper.postprocess import PostProcessor
from dotagent.fact_check.code.helper.execution import evaluate_test_cases_multi_solution
from dotagent.fact_check.utils.utils_json import CustomJSONEncoder

class code_pipeline(pipeline):
    def __init__(self, foundation_model, multi_solution_cnt, testcases_input_cnt):
        super().__init__('code', foundation_model)

        self.multi_solution_cnt = multi_solution_cnt
        self.testcases_input_cnt = testcases_input_cnt

        with open(os.path.join(self.prompts_path, "query_generation.yaml"), 'r') as file:
            data = yaml.load(file, Loader=yaml.FullLoader)
        self.query_generation_prompt = data['code']

    async def _testcases_input_generation(self, batch, testcases_input_cnt):
        messages_list = []
        if self.company == 'openai':
            messages_list = [
                [
                    {"role": "system", "content": self.query_generation_prompt['system']},
                    {"role": "user",
                     "content":
                         self.query_generation_prompt[
                             'user_testcases_' + str(testcases_input_cnt)
                             ].format(input_question=sample['prompt'],
                                      entry_point=sample['entry_point'])
                     },
                ]
                for sample in batch
            ]
        elif self.company == 'anthropic':
            messages_list = [self.query_generation_prompt[
                                 'user_testcases_' + str(testcases_input_cnt)
                                 ].format(input_question=sample['prompt'],
                                          entry_point=sample['entry_point'])
                             for sample in batch]
        return await self.chat.async_run(messages_list, Dict)
    
    async def _multi_solution_generation(self, batch, multi_solution_cnt):
        bsize = 15
        messages_list = [
            [
                {"role": "system", "content": self.query_generation_prompt['system']},
                {"role": "user", "content": self.query_generation_prompt[
                    'user_solutions'].format(input_question=sample['prompt'],
                                             entry_point=sample['entry_point'])},
            ]
            for sample in batch
        ]

        final_messages_list = [copy.deepcopy(messages)
                               for messages in messages_list
                               for _ in range(multi_solution_cnt)
                               ]

        responses = []
        for i in range(0, len(final_messages_list), bsize):
            batch = final_messages_list[i:i + bsize]
            responses += await self.chat.async_run(batch, Dict)
        
        # Split the list into lists of length of multi_solution_cnt
        responses_split = [responses[i:i + multi_solution_cnt]
                           for i in range(0, len(responses),
                                          multi_solution_cnt)]

        # Transform each element in each list
        multi_solutions = []
        for solutions in responses_split:
            key_names = [f"python_solution_{i}"
                         for i in range(1, multi_solution_cnt + 1)]
            new_element = {key: solutions[i]['python_solution']
            if solutions[i] != None else "None" for i, key in enumerate(key_names)}
            multi_solutions.append(new_element)

        return multi_solutions
    
    async def run_with_tool_live(self, batch, batch_size):
        testcases_input = await self._testcases_input_generation(batch, self.testcases_input_cnt)
        multi_solutions = await self._multi_solution_generation(batch, self.multi_solution_cnt)

        if testcases_input == None or multi_solutions == None:
            return None
        
        responses = []
        for i in range(batch_size):
            response = {'testcases_input': [],
                        'multi_solutions': [], 'with_tool_classification': "None"}
            try:
                response['testcases_input'] = list(testcases_input[i].values())
                # Append the solution to be verified to the LAST element
                # of multi_solutions
                response['multi_solutions']\
                    = [multi_solutions[i][f'python_solution_{j}']
                       for j in range(1, self.multi_solution_cnt + 1)] +\
                      [batch[i]['completion']]
            except:
                response['testcases_input'] = ["None"] * self.testcases_input_cnt
                response['multi_solutions'] = ["None"] * (self.multi_solution_cnt + 1)
            
            exec_result = evaluate_test_cases_multi_solution(
                batch[i]['prompt'], response['testcases_input'],
                response['multi_solutions'], timeout=0.1)
            response['exec_result'] = exec_result
            
            response['with_tool_classification'] = True
            # must pass all testcases to be classified as "True"
            for testcase_result in exec_result:
                # syntax or timeout error happening on the potential solution
                if isinstance(testcase_result[-1], str) \
                        and testcase_result[-1].startswith('FAILURE'):
                    response['with_tool_classification'] = False
                # majority voting. Note that the last element
                # is the solution to be verified. Also, multi solutions that return "FAILURE" are not counted and removed.
                else:
                    failure_indices = [
                        i for i, res in enumerate(testcase_result[:-1])
                        if isinstance(res, str) and res.startswith('FAILURE')]
                    testcase_result = [
                        res for i, res in enumerate(testcase_result)
                        if i not in failure_indices]

                    try:
                        if testcase_result[:-1].count(testcase_result[-1]) \
                                < math.ceil(len(testcase_result) / 2):
                            response['with_tool_classification'] = False
                    # sometimes numpy array is included in testcase_result, so this error will be raised
                    except:
                        response['with_tool_classification'] = False
            
            responses.append(response)

        return responses

    async def run_with_tool_api_call(self, prompts, responses, entry_points):

        # response preprocessing to extract the code snippet:
        claims = []
        for i, response in enumerate(responses):
            if "```python" in response:
                match = re.search(r"```python\n(.*?)\n```", response, re.DOTALL)
                if match:
                    claims.append(match.group(1))
                else:
                    claims.append("")
            elif "```" in response:
                match = re.search(r"```\n(.*?)\n```", response, re.DOTALL)
                if match:
                    claims.append(match.group(1))
                else:
                    claims.append("")
            else:
                claims.append(response)

        batch_size = 5
        num_batches = math.ceil(len(prompts) / batch_size)

        self.sample_list = [
            {"prompt": prompt, "response": response,
             "entry_point": entry_point, "completion": claim,
             "category": 'code'}
            for prompt, response, entry_point, claim
            in zip(prompts, responses, entry_points, claims)]

        for i in range(num_batches):
            print(i)
            batch_start = i * batch_size
            batch_end = min((i + 1) * batch_size, len(responses))

            responses_returned = await self.run_with_tool_live(self.sample_list[batch_start: batch_end], batch_end - batch_start)

            for j, response_returned in enumerate(responses_returned):
                index = batch_start + j
                self.sample_list[index].update({
                    'claim': self.sample_list[index]['completion'],
                    'testcases_queries': response_returned['testcases_input'],
                    'potential_solutions_queries': response_returned['multi_solutions'],
                    'exec_results': response_returned['exec_result'],
                    'claim_level_factuality': response_returned['with_tool_classification'],
                    'response_level_factuality': response_returned['with_tool_classification']
                })
                del self.sample_list[index]["completion"]

        return self.sample_list
        
    async def run_with_tool_dataset(self, annotated_dataset_path: str, with_tool_classified_dataset_path: str, rerun: bool = False, rerun_indices: list = []):
        data_path = with_tool_classified_dataset_path if rerun else annotated_dataset_path
        with open(data_path, 'r') as f:
            data = [json.loads(line) for line in f]
        self.sample_list = data
        rerun_elements = self.sample_list if not rerun else [self.sample_list[i] for i in rerun_indices]

        batch_size = 5
        num_batches = math.ceil(len(rerun_elements) / batch_size) # 5

        for i in range(num_batches):
            print(i)
            batch_start = i * batch_size
            batch_end = min((i + 1) * batch_size, len(rerun_elements))

            responses = await self.run_with_tool_live(rerun_elements[batch_start:batch_end], batch_end - batch_start)

            for j, response in enumerate(responses):
                index = batch_start + j if not rerun else rerun_indices[batch_start + j]
                self.sample_list[index]['with_tool_classification'] = response['with_tool_classification'] if response is not None else 'None'
                if response is not None:
                    self.sample_list[index].update({
                        'testcases_input': response['testcases_input'],
                        'multi_solutions': response['multi_solutions'],
                        'exec_result': response['exec_result']
                    })
        
            # save everything after each batch to prevent data loss
            with open(with_tool_classified_dataset_path, 'w') as f:
                for item in self.sample_list:
                    try:
                        json_str = json.dumps(item, cls=CustomJSONEncoder)
                    except:
                        continue
                    f.write(json_str + '\n')
    
    async def run_self_check_live(self, fewshot, batch):
        user_prompt_key = 'user_3_shot_CoT' if fewshot else 'user_zero_shot_CoT'
        messages_list = [
            [
                {"role": "system", "content": self.self_check_prompt['system']},
                {"role": "user", "content": self.self_check_prompt[user_prompt_key].format(input_question=response['prompt'], input_solution=response['completion'])},
            ]
            for response in batch
        ]
        return await self.chat.async_run(messages_list, Dict)

    async def run_self_check_dataset(self, annotated_dataset_path: str, self_check_classified_dataset_path: str, fewshot: bool = False, rerun: bool = False, rerun_indices: list = []):
        if rerun == False:
            with open(annotated_dataset_path, 'r') as f:
                self.sample_list = [json.loads(line) for line in f]
            rerun_elements = self.sample_list
        else:
            with open(self_check_classified_dataset_path, 'r') as f:
                self.sample_list = [json.loads(line) for line in f]
            rerun_elements = [self.sample_list[i] for i in rerun_indices]

        batch_size = 5
        num_batches = math.ceil(len(rerun_elements) / batch_size)

        for i in range(num_batches):
            print(i)
            batch_start = i * batch_size
            batch_end = (i + 1) * batch_size
            batch = rerun_elements[batch_start:batch_end]

            responses = await self.run_self_check_live(fewshot, batch)
            for j, response in enumerate(responses):
                index = batch_start + j if rerun == False else rerun_indices[batch_start + j]
                self.sample_list[index]['self_check_classification'] = response.get('factuality', 'None') if response is not None else 'None'
                self.sample_list[index]['self_check_reasoning'] = response.get('reasoning', 'None') if response is not None else 'None'
        
            # save everything after each batch to prevent data loss
            with open(self_check_classified_dataset_path, 'w') as f:
                for item in self.sample_list:
                    json_str = json.dumps(item)
                    f.write(json_str + '\n')


"""Google Calendar toolkit."""

# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dotagent.tools.toolkits.base import BaseToolkit
from dotagent.tools.basetool import BaseTool
from dotagent.tools.toolkits.google_calendar_toolkit.google_calendar.base import LoadData, CreateEvent, GetDate
from typing import List

SCOPES = ["https://www.googleapis.com/auth/calendar"]


class GoogleCalendarToolkit(BaseToolkit):
    """Google Calendar toolkit.

    Currently a simple wrapper around the data loader.
    TODO: add more methods to the Google Calendar toolkit.

    """

    def get_tools(self) -> List[BaseTool]:
        """Get the tools in the toolkit."""
        return [
            LoadData(),
            CreateEvent(),
            GetDate(),   
        ]

    


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

from collections import defaultdict

from dotagent.fact_check.code.helper.io_utils import Tools

STOP_TOKEN = ['\nclass', '\ndef', '\n#', '\nif', '\nprint']

class PostProcessor:
    @staticmethod
    def map_task_id_for_solution(predict_path, source_path):
        database = dict()
        raw_problems = Tools.load_tasks(source_path)
        for task_id in raw_problems.keys():
            database[raw_problems[task_id]['prompt']] = raw_problems[task_id]

        result = []
        predictions = Tools.load_jsonl(predict_path)
        
        for pre in predictions:
            task = database[pre['prompt']]
            
            for sample in pre['samples']:
                processed_code = PostProcessor.solution_extract(sample)
                result.append({
                    'task_id': task['task_id'],
                    'prompt': pre['prompt'],
                    'test': task['test'],
                    'entry_point': task['entry_point'],
                    'completion': processed_code
                })
        return result, len(raw_problems)

    @staticmethod
    def solution_extract(content):
        for identifier in STOP_TOKEN:
            if identifier in content:
                content = content.split(identifier)[0]
        return content



from .._transformers import Transformers

class StableLMChat(Transformers):
    """ A HuggingFace transformers version of the StableLM language model with Compiler support.
    """

    llm_name: str = "stablelm"

    @staticmethod
    def role_start(role):
        return "<|"+role.upper()+"|>"
    
    @staticmethod
    def role_end(role):
        return '' 


from typing import Type, Optional
from pydantic import BaseModel, Field
from requests.exceptions import RequestException
from bs4 import BeautifulSoup
import re
import requests
import random


USER_AGENTS = [
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3",
    #... Rest of the user agents...
]

class WebScraperSchema(BaseModel):
    website_url: str = Field(
        ...,
        description="Valid website url without any quotes.",
    )

class WebScraperTool(BaseTool):
    """
    Web Scraper tool

    Attributes:
        name : The name.
        description : The description.
        args_schema : The args schema.
    """
    name = "WebScraperTool"
    description = (
        "Used to scrape website urls and extract text content"
    )
    args_schema: Type[WebScraperSchema] = WebScraperSchema

    def run(self, tool_input: str, **kwargs: Any) -> Any:
        """
        Execute the Web Scraper tool.

        Args:
            tool_input : The website url to scrape.

        Returns:
            The text content of the website.
        """
        content = self.extract_with_bs4(tool_input)
        max_length = len(' '.join(content.split(" ")[:600]))
        return content[:max_length]

    def extract_with_bs4(self, url):
        headers = {
            "User-Agent": random.choice(USER_AGENTS)
        }
        try:
            response = requests.get(url, headers=headers, timeout=10)
            if response.status_code == 200:
                soup = BeautifulSoup(response.text, 'html.parser')
                for tag in soup(['script', 'style', 'nav', 'footer', 'head', 'link', 'meta', 'noscript']):
                    tag.decompose()

                main_content_areas = soup.find_all(['main', 'article', 'section', 'div'])
                if main_content_areas:
                    main_content = max(main_content_areas, key=lambda x: len(x.text))
                    content_tags = ['p', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6']
                    content = ' '.join([tag.text.strip() for tag in main_content.find_all(content_tags)])
                else:
                    content = ' '.join([tag.text.strip() for tag in soup.find_all(['p', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6'])])

                content = re.sub(r'\t', ' ', content)
                content = re.sub(r'\s+', ' ', content)
                return content
            elif response.status_code == 404:
                return f"Error: 404. Url is invalid or does not exist. Try with valid url..."
            else:
                logger.error(f"Error while extracting text from HTML (bs4): {response.status_code}")
                return f"Error while extracting text from HTML (bs4): {response.status_code}"

        except Exception as e:
            logger.error(f"Unknown error while extracting text from HTML (bs4): {str(e)}")
            return ""



from dotagent.llms._openai import OpenAI
from dotagent import compiler

llm = OpenAI(model="gpt-4")

prompt = """
{{#system~}}
You are a helpful and terse assistant. 
{{~/system}}

{{#user~}}
Given a chunk of text, provide an informative summary in not more than {{max_words}} words.

{{input_text}}
{{~/user}}

{{#assistant~}}
{{gen 'response'}}
{{~/assistant}}

"""

engine = compiler(prompt, llm = llm, silent = True)

class Summarizer:

    def __init__(
            self,
            engine = engine
        ) -> None:
        
        self.engine = engine


    def summarize(self, input_text, max_words):

        result = self.engine(input_text = input_text, max_words = max_words)

        response = result['response']
        
        return response








"""Deepdoctection Data Reader."""

from pathlib import Path
from typing import Optional, Set
from typing import Dict, List
from dotagent.knowledgebase.document_loaders.basereader import BaseReader
from dotagent.schema import DocumentNode


class DeepDoctectionReader(BaseReader):
    """Deepdoctection reader for pdf's.

    Uses deepdoctection as a library to parse PDF files.

    """

    def __init__(self, attrs_as_metadata: Optional[Set] = None) -> None:
        """Init params."""
        import deepdoctection as dd

        self.analyzer = dd.get_dd_analyzer()
        self.attrs_as_metadata = attrs_as_metadata or set()

    def load_data(
        self, file: Path, extra_info: Optional[Dict] = None
    ) -> List[DocumentNode]:
        """Parse file."""
        df = self.analyzer.analyze(path=str(file))
        df.reset_state()
        doc = iter(df)
        result_docs = []
        for page in doc:
            doc_text = page.text
            extra_info = {
                k: getattr(page, k) for k in self.attrs_as_metadata if hasattr(page, k)
            }
            result_docs.append(DocumentNode(text=doc_text, extra_info=extra_info))
        return result_docs



"""Notion tool spec."""

#from llama_index.tools.tool_spec.base import BaseToolSpec
from dotagent.tools.toolkits.notion_toolkit.notion.utils import NotionPageReader
from dotagent.tools.basetool import BaseTool
from typing import Optional, List, Type, Dict, Any
import requests
from pydantic import BaseModel, Field

SEARCH_URL = "https://api.notion.com/v1/search"


class NotionLoadDataSchema(BaseModel):
    """Notion load data schema."""

    page_ids: Optional[List[str]] = None
    database_id: Optional[str] = None


class NotionSearchDataSchema(BaseModel):
    """Notion search data schema."""

    query: str
    direction: Optional[str] = None
    timestamp: Optional[str] = None
    value: Optional[str] = None
    property: Optional[str] = None
    page_size: int = 100

class NotionBase(BaseTool):

    def __init__(self, integration_token: Optional[str] = None) -> None:
        """Initialize with parameters."""
        self.reader = NotionPageReader(integration_token=integration_token)


class LoadDataArgsSchema(BaseModel):
    page_ids: Optional[List[str]] = Field(
        ...,
        description="******Provide Description About PageId*******"
    )
    database_id: Optional[str] = Field(
        ...,
        description="******Description about this Parameter********"
    )

class SearchDataArgsSchema(BaseModel):
    query: str = Field(
        ...,
        description="Info about parameter"
    )
    direction: Optional[str] = Field(
        ...,
        description="Info about parameter"
    )
    timestamp: Optional[str] = Field(
        ...,
        description="Info about parameter"
    )
    value: Optional[str] = Field(
        ...,
        description="Info about parameter"
    )
    property: Optional[str] = Field(
        ...,
        description="Info about parameter"
    )
    page_size: int = Field(
        ...,
        description="Info about parameter"
    )

class LoadData(NotionBase):
    name : str = "Load Data"
    description : str = "Loads content from a set of page ids or a database id."
    args_schema : Type[LoadDataArgsSchema] = LoadDataArgsSchema

    def load_data(
        self, page_ids: Optional[List[str]] = None, database_id: Optional[str] = None
    ) -> str:
        """Loads content from a set of page ids or a database id.

        Don't use this endpoint if you don't know the page ids or database id.

        """
        page_ids = page_ids or []
        docs = self.reader.load_data(page_ids=page_ids, database_id=database_id)
        return "\n".join([doc.get_content() for doc in docs])
    
    def run(
        self, page_ids: Optional[List[str]] = None, database_id: Optional[str] = None
    ) -> str:
        try:
            return self.load_data(self, page_ids=page_ids, database_id=database_id)
        except Exception as e:
            return e

    
class SearchData(NotionBase):
    name: str = "Search Data"
    description: str = "Search a list of relevant pages.Contains metadata for each page (but not the page content)"
    args_schema: Type[SearchDataArgsSchema] = SearchDataArgsSchema
    def search_data(
        self,
        query: str,
        direction: Optional[str] = None,
        timestamp: Optional[str] = None,
        value: Optional[str] = None,
        property: Optional[str] = None,
        page_size: int = 100,
    ) -> str:
        """Search a list of relevant pages.

        Contains metadata for each page (but not the page content).

        """
        payload: Dict[str, Any] = {
            "query": query,
            "page_size": page_size,
        }
        if direction is not None or timestamp is not None:
            payload["sort"] = {}
            if direction is not None:
                payload["sort"]["direction"] = direction
            if timestamp is not None:
                payload["sort"]["timestamp"] = timestamp

        if value is not None or property is not None:
            payload["filter"] = {}
            if value is not None:
                payload["filter"]["value"] = value
            if property is not None:
                payload["filter"]["property"] = property

        response = requests.post(SEARCH_URL, json=payload, headers=self.reader.headers)
        response_json = response.json()
        response_results = response_json["results"]
        return response_results
    
    def run(
        self,
        query: str,
        direction: Optional[str] = None,
        timestamp: Optional[str] = None,
        value: Optional[str] = None,
        property: Optional[str] = None,
        page_size: int = 100,
    ) -> str:
        try:
            return self.search_data(
                query=query,
                direction=direction,
                timestamp=timestamp,
                value=value,
                property=property,
                page_size=page_size
            )
        except Exception as e:
            return e


"""Wrappers around embedding modules."""
import logging
from typing import Any

from dotagent.vectorstores.embeddings.aleph_alpha import (
    AlephAlphaAsymmetricSemanticEmbedding,
    AlephAlphaSymmetricSemanticEmbedding,
)
from dotagent.vectorstores.embeddings.bedrock import BedrockEmbeddings
from dotagent.vectorstores.embeddings.cohere import CohereEmbeddings
from dotagent.vectorstores.embeddings.dashscope import DashScopeEmbeddings
from dotagent.vectorstores.embeddings.deepinfra import DeepInfraEmbeddings
from dotagent.vectorstores.embeddings.elasticsearch import ElasticsearchEmbeddings
from dotagent.vectorstores.embeddings.embaas import EmbaasEmbeddings
from dotagent.vectorstores.embeddings.fake import FakeEmbeddings
from dotagent.vectorstores.embeddings.google_palm import GooglePalmEmbeddings
from dotagent.vectorstores.embeddings.huggingface import (
    HuggingFaceSetenceTransformersEmbeddings,
    HuggingFaceInstructEmbeddings,
    HuggingFaceHubEmbeddings,
)
from dotagent.vectorstores.embeddings.jina import JinaEmbeddings
from dotagent.vectorstores.embeddings.llamacpp import LlamaCppEmbeddings
from dotagent.vectorstores.embeddings.minimax import MiniMaxEmbeddings
from dotagent.vectorstores.embeddings.modelscopehub import ModelScopeEmbeddings
from dotagent.vectorstores.embeddings.mosaicml import MosaicMLInstructorEmbeddings
from dotagent.vectorstores.embeddings.openai import OpenAIEmbeddings
from dotagent.vectorstores.embeddings.tensorflowhub import TensorflowHubEmbeddings

logger = logging.getLogger(__name__)

__all__ = [
    "AlephAlphaAsymmetricSemanticEmbedding",
    "AlephAlphaSymmetricSemanticEmbedding",
    "BedrockEmbeddings",
    "CohereEmbeddings",
    "DashScopeEmbeddings",
    "DeepInfraEmbeddings",
    "ElasticsearchEmbeddings",
    "EmbaasEmbeddings",
    "FakeEmbeddings",
    "GooglePalmEmbeddings",
    "HuggingFaceSetenceTransformersEmbeddings",
    "HuggingFaceInstructEmbeddings",
    "HuggingFaceHubEmbeddings",
    "JinaEmbeddings",
    "LlamaCppEmbeddings",
    "MiniMaxEmbeddings",
    "ModelScopeEmbeddings",
    "MosaicMLInstructorEmbeddings",
    "OpenAIEmbeddings",
    "TensorflowHubEmbeddings",
]




from .._utils import ContentCapture

async def block(name=None, hidden=False, _parser_context=None):
    ''' Generic block-level element.

    This is useful for naming or hiding blocks of content.

    Parameters
    ----------
    name : str
        The name of the block. A variable with this name will be set with the generated block content.
    hidden : bool
        Whether to include the generated block content in future LLM context.
    '''

    parser = _parser_context['parser']
    variable_stack = _parser_context['variable_stack']
    
    # capture the content of the block
    with ContentCapture(variable_stack, hidden) as new_content:

        # visit the block content
        new_content += await parser.visit(
            _parser_context['block_content'][0],
            variable_stack,
            next_node=_parser_context["next_node"],
            next_next_node=_parser_context["next_next_node"],
            prev_node=_parser_context["prev_node"]
        )

        # set the variable if needed
        if name is not None:
            variable_value = str(new_content)
            variable_stack[name] = variable_value
block.is_block = True